import gempipe


from .fluxbal import fba_no_warnings




def apply_medium_given_column(logger, model, medium, column, is_reference=False):
        
        
    # retrieve metadata
    description = column.iloc[0]
    doi = column.iloc[1]
    author = column.iloc[2]
    units = column.iloc[3]

    
    # convert substrates to dict
    column = column.iloc[4:]
    column = column.to_dict()

    
    # add default elements (acqueous media)
    column['h2o'] = 'NL'
    column['h'] = '-0.0001'  # pH=7


    # reset exchanges
    gempipe.reset_growth_env(model)    
    modeled_rids = [r.id for r in model.reactions]


    for substrate, value in column.items():

        if type(value)==float:
            continue   # empty cell, exchange will remain close

            
        # check if exchange is modeled
        if is_reference == False: 
            if f'EX_{substrate}_e' not in modeled_rids:
                logger.error(f"No exchange reaction found for substrate '{substrate}' in medium '{medium}'.")
                return 1
        else:  # external reference models might follow different standards.
            # The exr might not be present. 
            if f'EX_{substrate}_e' not in modeled_rids:
                logger.info(f"Reference has no exchange reaction for '{substrate}' in medium '{medium}': this substrate will be ignored.")
                continue


        # case "not limiting"
        value = value.strip().rstrip()
        if value == 'NL':   # non-limiting case
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = -1000


        # case "single value"
        elif '+-' not in value and '±' not in value:  # single number case
            value = value.replace(' ', '')  # eg "- 0.03" --> "-0.03"
            try: value = float(value)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value}.")
                return 1
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = value


        # case "with experimental error"
        else:  # value +- error
            if '±' in value: 
                value, error = value.split('±', 1)
            else: value, error = value.split('+-', 1)
            value = value.rstrip()
            error = error.strip()
            value = value.replace(' ', '')  # eg "- 0.03" --> "-0.03"
            try: value = float(value)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value} +- {error}.")
                return 1
            try: error = float(error)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value} +- {error}.")
                return 1
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = value -error
            model.reactions.get_by_id(f'EX_{substrate}_e').upper_bound = value +error

    return 0



def initialize_model(logger, model, dbexp, initialize, media):
        
        
    if initialize in ['None', 'none']:
        logger.info(f"Initialization will be skipped.")
        return 0
    elif initialize == '-':
        if media == '-':
            logger.info(f"No media provided: initialization will be skipped.")
            return 0
        else:
            media = media.split(',')
            medium = media[0]  # taking the first medium
    else:
        medium = initialize


    if medium not in dbexp['media'].columns:
        logger.warning(f"Medium '{medium}' does not exists: initialization will be skipped.")
        return 0


    response = apply_medium_given_column(logger, model, medium, dbexp['media'][medium])
    if response == 1: return 1
    res, obj_value, status = fba_no_warnings(model)
    logger.info(f"Initialized on medium '{medium}': {obj_value} ({status})")


    return 0


import pandas as pnd
import cobra


from .repeating import check_author
from .repeating import check_rstring_arrow
from .repeating import check_gpr
from .repeating import add_reaction

from .manual import get_manual_sinks
from .manual import get_manual_demands


    
def introduce_metabolites(logger, db, model, idcollection_dict, kegg_compound_to_others, goodbefore, onlyauthor):
    goodbefore_reached = False

    
    logger.debug("Checking duplicated metabolite IDs...")
    if len(set(db['M']['pure_mid'].to_list())) != len(db['M']): 
        pure_mids = db['M']['pure_mid'].to_list()
        duplicates = list(set([item for item in pure_mids if pure_mids.count(item) > 1]))
        logger.error(f"Sheet 'M' has duplicated metabolites: {duplicates}.")
        return 1
   
        
    # parse M:
    logger.debug("Parsing metabolites...")
    db['M'] = db['M'].set_index('pure_mid', drop=True, verify_integrity=True)
    kc_ids_modeled = set()   # account for kc codes modeled
    for pure_mid, row in db['M'].iterrows():
        
        
        # skip empty lines!
        if type(pure_mid) != str: continue
        if pure_mid.strip() == '': continue
        if pure_mid == goodbefore:
            goodbefore_reached = True
            
            
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor == None:
                logger.info(f"Skipping metabolite '{pure_mid}' as requested with --goodbefore[0] '{goodbefore}'.")
                continue
        
        
        # parse author
        response = check_author(logger, pure_mid, row, db, 'M')
        if type(response) == int: return 1
        else: authors = response
        
        
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor != None and onlyauthor not in authors:
                authors_string = '; '.join(authors)
                logger.info(f"Skipping metabolite '{pure_mid}' (authors '{authors_string}') as requested with --goodbefore[0] '{goodbefore}' and --onlyauthor '{onlyauthor}'.")
                continue
            
        
        # parse formula:
        if pnd.isna(row['formula']):
            logger.error(f"Metabolite '{pure_mid}' has missing formula: '{row['formula']}'.")
            return 1
  
        
        # parse charge: 
        if pnd.isna(row['charge']): 
            logger.error(f"Metabolite '{pure_mid}' has missing charge: '{row['charge']}'.")
            return 1
        
        
        # check if 'kc' codes are real:
        if pnd.isna(row['kc']): 
            logger.error(f"Metabolite '{pure_mid}' has missing KEGG annotation (kc): '{row['kc']}'.")
            return 1
        kc_ids = row['kc'].split(';')
        kc_ids = [i.strip() for i in kc_ids]
        for kc_id in kc_ids:
            if kc_id == 'CXXXXX':  # not in KEGG; could be knowledge gap (e.g. methyl group acceptor in R10404)
                logger.debug(f"Metabolite '{pure_mid}' is not in KEGG ('{kc_id}')!")
                continue  
            if kc_id not in idcollection_dict['kc']:
                logger.error(f"Metabolite '{pure_mid}' has invalid KEGG annotation (kc): '{kc_id}'.")
                return 1
            if kc_id in kc_ids_modeled:
                logger.error(f"KEGG annotation (kc) '{kc_id}' used in metabolite '{pure_mid}' is duplicated.")
                return 1
            if kc_id != 'CXXXXX':
                kc_ids_modeled.add(kc_id)
            
            
        # check the existance of the inchikey
        if pnd.isna(row['inchikey']): 
            logger.error(f"Metabolite '{pure_mid}' has missing inchikey: '{row['inchikey']}'.")
            return 1
        # check inchikey format:
        if len(row['inchikey']) != 27 or row['inchikey'][14] != '-' or row['inchikey'][25] != '-':
            logger.error(f"Metabolite '{pure_mid}' has badly formatted inchikey: '{row['inchikey']}'.")
            return 1
        
        
        # check if this 'kc' is already in BiGG (rely on MNX)
        eqbiggids = set()
        for kc_id in kc_ids:
            if kc_id != 'CXXXXX':
                if kc_id in kegg_compound_to_others.keys():
                    for eqbiggid in kegg_compound_to_others[kc_id]['bigg.metabolite']:
                        eqbiggids.add(eqbiggid)
        if pure_mid not in eqbiggids and eqbiggids != set():
            logger.debug(f"Metabolites '{'; '.join(kc_ids)}' already in BiGG as {eqbiggids} ({authors} gave '{pure_mid}').")        
        
        
        # add metabolite to model
        m = cobra.Metabolite(f'{pure_mid}_c')
        model.add_metabolites([m])
        m = model.metabolites.get_by_id(f'{pure_mid}_c')
        m.name = row['name'].strip()
        m.formula = row['formula']
        m.charge = row['charge']
        m.compartment='c'
        
        
        # add annotations to model (same order of Memote)
        ankeys = [
            'pubchem.compound', 'kegg.compound', 'seed.compound',
            'inchikey', 'inchi', 'chebi', 'hmdb', 'reactome',
            'metanetx.chemical', 'bigg.metabolite', 'biocyc',
        ]
        # initialize sets:
        for ankey in ankeys:
            if ankey == 'kegg.compound': m.annotation[ankey] = set(kc_ids) - set(['CXXXXX'])
            else: m.annotation[ankey] = set()
        # populate sets:
        for kc_id in kc_ids:
            if kc_id != 'CXXXXX':
                if kc_id in kegg_compound_to_others.keys():
                    for ankey in ankeys:
                        m.annotation[ankey].update(kegg_compound_to_others[kc_id][ankey])
        # save as list: 
        for ankey in ankeys:
            m.annotation[ankey] = list(m.annotation[ankey])
            
            
        # replace inchikey with manually-curated
        if m.annotation['inchikey'] != [] and m.annotation['inchikey'] != [row['inchikey']]:
            logger.debug(f"Metabolite '{pure_mid}': manual-curated inchikey ({[row['inchikey']]}) is diferent from the one derived from MNX ({m.annotation['inchikey']}).")
        m.annotation['inchikey'] = [row['inchikey']]  # force the manual-curated version
        if m.annotation['inchikey'] == ['XXXXXXXXXXXXXX-XXXXXXXXXX-X']:
            m.annotation['inchikey'] = []
            
            
        # add SBO annotation
        m.annotation['sbo'] = ['SBO:0000247']  # generic metabolite
   
        
    
    if goodbefore != None and goodbefore_reached == False:
        logger.info(f"Metabolite '{goodbefore}' never reached. Are you sure about your --goodbefore?")
                    
                    
    return model
    
    
    
def introduce_reactions(logger, db, model, idcollection_dict, kegg_reaction_to_others, goodbefore, onlyauthor): 
    goodbefore_reached = False
                    
    
    logger.debug("Checking duplicated reaction IDs...")
    if len(set(db['R']['rid'].to_list())) != len(db['R']): 
        pure_mids = db['R']['rid'].to_list()
        duplicates = list(set([item for item in pure_mids if pure_mids.count(item) > 1]))
        logger.error(f"Sheet 'R' has duplicated reactions: {duplicates}.")
        return 1
    
        
    # parse R:
    logger.debug("Parsing reactions...")
    db['R'] = db['R'].set_index('rid', drop=True, verify_integrity=True)
    for rid, row in db['R'].iterrows():
        
        
        # skip empty lines!
        if type(rid) != str: continue
        if rid.strip() == '': continue
        if rid == goodbefore:
            goodbefore_reached = True
        
        
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor == None:
                logger.info(f"Skipping reaction '{rid}' as requested with --goodbefore[1] '{goodbefore}'.")
                continue
        
        
        # parse author
        response = check_author(logger, rid, row, db, 'R')
        if type(response) == int: return 1
        else: authors = response
                    
                    
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor != None and onlyauthor not in authors:
                authors_string = '; '.join(authors)
                logger.info(f"Skipping reaction '{rid}' (authors '{authors_string}') as requested with --goodbefore[1] '{goodbefore}' and --onlyauthor '{onlyauthor}'.")
                continue
        
        
        # parse reaction string
        response = check_rstring_arrow(logger, rid, row, 'R')
        if response == 1: return 1
        

        # check if 'kr' codes are real:
        if pnd.isna(row['kr']): 
            logger.error(f"Reaction '{rid}' has missing KEGG annotation (kr): '{row['kr']}'.")
            return 1
        kr_ids = row['kr'].split(';')
        kr_ids = [i.strip() for i in kr_ids]
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX':  # not in KEGG; could be knowledge gap 
                logger.debug(f"Reaction '{rid}' is not in KEGG ('{kr_id}')!")
                continue  
            if kr_id not in idcollection_dict['kr']:
                logger.error(f"Reaction '{rid}' has invalid KEGG annotation (kr): '{kr_id}'.")
                return 1
        
            
        # check GPR:
        response = check_gpr(logger, rid, row, kr_ids, idcollection_dict, 'R')
        if response == 1: return 1
    
    
        # check if this 'kr' is already in BiGG (rely on MNX)
        eqbiggids = set()
        for kr_id in kr_ids:
            if kr_id != 'RXXXXX':
                if kr_id in kegg_reaction_to_others.keys():
                    for eqbiggid in kegg_reaction_to_others[kr_id]['bigg.reaction']:
                        eqbiggids.add(eqbiggid)
        if rid not in eqbiggids and eqbiggids != set():
            logger.debug(f"Reactions '{'; '.join(kr_ids)}' already in BiGG as {eqbiggids} ({authors} gave '{rid}').") 
        
        
        # add reaction to model
        response = add_reaction(logger, model, rid, row, kr_ids, kegg_reaction_to_others, 'R')
        if response == 1: return 1
               
    
    if goodbefore != None and goodbefore_reached == False:
        logger.info(f"Reaction '{goodbefore}' never reached. Are you sure about your --goodbefore?")
    
    
    return model
      
    
    
def introduce_transporters(logger, db, model, idcollection_dict, kegg_reaction_to_others, goodbefore, onlyauthor): 
    goodbefore_reached = False
                    
    
    
    def clone_to_external(model, mid_c, mid_e):
    
        m = cobra.Metabolite(f'{mid_e}')
        model.add_metabolites([m])
        
        m_c = model.metabolites.get_by_id(f'{mid_c}')
        m_e = model.metabolites.get_by_id(f'{mid_e}')
        m_e.compartment='e'
        
        m_e.name = m_c.name
        m_e.formula = m_c.formula
        m_e.charge = m_c.charge
        
        m_e.annotation = m_c.annotation   # transfer all annotations, including SBO!
            
    
    def add_exchange_reaction(model, mid_e):
        
        r = cobra.Reaction(f'EX_{mid_e}')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'EX_{mid_e}')
        r.name = f"Exchange for {model.metabolites.get_by_id(mid_e).name}"
        r.build_reaction_from_string(f'{mid_e} --> ')
        r.bounds = (0, 1000)
            
        # add SBO annotation
        r.annotation['sbo'] = ['SBO:0000627']  # exchange reaction    
        
    
    
    
    # get all already inserted metabolites
    mids_parsed = [m.id for m in model.metabolites]
    rids_parsed = [r.id for r in model.reactions]
    
    
    # protons may not have an explicit transporter
    clone_to_external(model, 'h_c', 'h_e')
    mids_parsed.append('h_e')
    add_exchange_reaction(model, 'h_e')
    rids_parsed.append(f'EX_h_e')
    
    
    # parse T:
    logger.debug("Parsing transporters...")
    db['T'] = db['T'].set_index('rid', drop=True, verify_integrity=True)
    for rid, row in db['T'].iterrows():
        
        
        # skip empty lines!
        if type(rid) != str: continue
        if rid.strip() == '': continue
        if rid == goodbefore:
            goodbefore_reached = True
            
            
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor == None:
                logger.info(f"Skipping transport '{rid}' as requested with --goodbefore[2] '{goodbefore}'.")
                continue
        
        
        # parse author
        response = check_author(logger, rid, row, db, 'T')
        if type(response) == int: return 1
        else: authors = response
                    
                    
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor != None and onlyauthor not in authors:
                authors_string = '; '.join(authors)
                logger.info(f"Skipping transport '{rid}' (authors '{authors_string}') as requested with --goodbefore[2] '{goodbefore}' and --onlyauthor '{onlyauthor}'.")
                continue
        
        
        # parse reaction string
        response = check_rstring_arrow(logger, rid, row, 'T')
        if response == 1: return 1
    
    
        # check if 'kr' codes are real:
        if pnd.isna(row['kr']): 
            logger.error(f"Reaction '{rid}' has missing KEGG annotation (kr): '{row['kr']}'.")
            return 1
        if row['kr'] != '-':
            kr_ids = row['kr'].split(';')
            kr_ids = [i.strip() for i in kr_ids]
            for kr_id in kr_ids:
                if kr_id not in idcollection_dict['kr']:
                    logger.error(f"Reaction '{rid}' has invalid KEGG annotation (kr): '{kr_id}'.")
                    return 1
        else: kr_ids = []

            
        # check GPR:
        response = check_gpr(logger, rid, row, kr_ids, idcollection_dict, 'T')
        if response == 1: return 1
        
        
        # get involved metabolites:
        involved_mids = row['rstring'].split(' ')
        involved_mids = [i for i in involved_mids if i not in ['-->', '<=>']]
        
        
        # the external metabolite must be already modeled as cytosolic
        for mid in involved_mids: 
            if mid.endswith('_e'):
                mid_e = mid
                mid_c = mid.rsplit('_', 1)[0] + '_c'
                if mid_c not in mids_parsed:
                    logger.error(f"{rid}: the metabolite '{mid_c}', counterpart of '{mid_e}', was not previously modeled.")
                    return 1
                
                # add external metabolite to model
                if mid_e not in mids_parsed:
                    clone_to_external(model, mid_c, mid_e)
                    mids_parsed.append(mid_e)
                    
                # add exchange reaction to model
                if f'EX_{mid_e}' not in rids_parsed:
                    add_exchange_reaction(model, mid_e)
                    rids_parsed.append(f'EX_{mid_e}')
                    
                    
        # check if this 'kr' is already in BiGG (rely on MNX)
        eqbiggids = set()
        for kr_id in kr_ids:
            if kr_id != '-':  # (was 'RXXXXX' for metabolic reactions)
                if kr_id in kegg_reaction_to_others.keys():
                    for eqbiggid in kegg_reaction_to_others[kr_id]['bigg.reaction']:
                        eqbiggids.add(eqbiggid)
        if rid not in eqbiggids and eqbiggids != set():
            logger.debug(f"Reactions '{'; '.join(kr_ids)}' already in BiGG as {eqbiggids} ({authors} gave '{rid}').") 
                    
                    
        # add reaction to model
        response = add_reaction(logger, model, rid, row, kr_ids, kegg_reaction_to_others, 'T')
        if response == 1: return 1
        
        
    if goodbefore != None and goodbefore_reached == False:
        logger.info(f"Transport '{goodbefore}' never reached. Are you sure about your --goodbefore?")
        
    
    return model



def introduce_sinks_demands(logger, model): 
    
    sinks = get_manual_sinks()
    demands = get_manual_demands()
    
    
    for puremid in sinks: 
        r = cobra.Reaction(f'sn_{puremid}_c')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'sn_{puremid}_c')
        r.name = f"Sink for {model.metabolites.get_by_id(f'{puremid}_c').name}"
        r.build_reaction_from_string(f'{puremid}_c <=> ')
        r.bounds = (-1000, 1000)
        
        # add SBO annotation
        r.annotation['sbo'] = ['SBO:0000632']  # sink reaction 
    
    
    for puremid in demands: 
        r = cobra.Reaction(f'dm_{puremid}_c')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'dm_{puremid}_c')
        r.name = f"Demand for {model.metabolites.get_by_id(f'{puremid}_c').name}"
        r.build_reaction_from_string(f'{puremid}_c --> ')
        r.bounds = (0, 1000)
        
        # add SBO annotation
        r.annotation['sbo'] = ['SBO:0000628']  # demand reaction 
    
    
    return model



        

    



import pandas as pnd
import gempipe


from ..commons import apply_medium_given_column



def biosynthesis_on_media(logger, model, dbexp, media, biosynth):
    
    
    # check if requested
    df_S = pnd.DataFrame()
    if not biosynth:
        return df_S
    
    
    # prepare sheet for excel output
    df_S['name'] = None
    for m in model.metabolites:
        if m.id.endswith('_c'):
            df_S.loc[m.id, 'name'] = m.name
            
            
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: metabolites biosynthesis check will be skipped.")
        return df_S
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in dbexp['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(dbexp['media'].columns)}.")
        return 1
    
    
    logger.info(f"Checking metabolites biosynthesis on {len(media)} media...")
    for medium in media:
        if medium not in dbexp['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_S[f'{medium}'] = None
        logger.debug(f"Checking metabolites biosynthesis on medium '{medium}'...")
            
            
        # apply medium :
        response = apply_medium_given_column(logger, model, medium, dbexp['media'][medium])
        if response == 1: return 1
            
    
        for m in model.metabolites:
            if m.id.endswith('_c'):
                dem = '/'
                binary, obj_value, status = gempipe.can_synth(model, m.id)


                # check if it's dead-end metabolite (dem)
                if binary == False:
                    dem = False
                    is_consumed = False
                    is_produced = False
                    for r in m.reactions:
                        if m.id in [m2.id for m2 in r.reactants]:
                            is_consumed = True
                        if m.id in [m2.id for m2 in r.products]:
                            is_produced = True
                    if   is_consumed and not is_produced:
                        dem = 'no_production'
                    elif is_produced and not is_consumed:
                        dem = 'no_consumption' 

                
                if status == 'infeasible': 
                    df_S.loc[m.id, f'{medium}'] = 'infeasible'
                elif obj_value == 0:
                    if dem == '/':
                        df_S.loc[m.id, f'{medium}'] = 'blocked'
                    else:
                        df_S.loc[m.id, f'{medium}'] = dem
                else:
                    df_S.loc[m.id, f'{medium}'] = obj_value
            
    
    return df_S




import warnings



def log_metrics(logger, model):
    
    
    G = len([g.id for g in model.genes])
    R = len([r.id for r in model.reactions if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))==1 and len(r.metabolites)!=1 and r.id!='Biomass'])
    T = len([r.id for r in model.reactions if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))!=1 and len(r.metabolites)!=1 and r.id!='Biomass'])
    A = len([r.id for r in model.reactions if len(r.metabolites)==1 or r.id=='Biomass'])
    M = len([m.id for m in model.metabolites])
    uM = len(set([m.id.rsplit('_',1)[0] for m in model.metabolites]))
    gr = len([gr.id for gr in model.groups])
    bP = len(set([m.id for m in model.reactions.get_by_id('Biomass').reactants]) - set(['atp_c', 'h2o_c']))

    
    logger.info(f"Done! [G: {G}, R: {R}, T: {T}, A: {A}, uM: {uM}, bP: {bP}]")
    
    
    
def log_unbalances(logger, model):
    
    
    balance_threshold = 1e-10
    
    cnt_mass = 0
    cnt_charge = 0
    for r in model.reactions: 
        if len(r.metabolites) != 1 and r.id != 'Biomass':   # exclude artificials
            
            
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # avoids warnings like "UserWarning: 9.823 is not an integer (in formula C9.823H5.5789999999999935N3.6769999999999996O3.628P1.0)"

                
                check = r.check_mass_balance()
                for key, value in check.items():
                    if abs(value) > balance_threshold:

                        if key == 'charge':
                            cnt_charge += 1
                            break
                        else:
                            cnt_mass += 1
                            break
    
    logger.info(f"Found {cnt_mass} mass and {cnt_charge} charge unbalances >{balance_threshold}.")
        
        
        
def show_contributions(logger, db, goodbefore):
    
    
    if goodbefore != [None, None, None]:
        logger.debug(f"Contributions  counter disabled when using --goodbefore.")
        return 0
    
    
    # create a counter for each author
    cnt = {author: 0 for author in db['authors']['username']}
    cnt_tot = 0
    
    
    for index, row in db['R'].iterrows():
        if type(row['author']) != str: 
            logger.error(f"Missing author in tab 'R', rid '{row['rid']}'.")
            return 1
        for author in row['author'].split(';'):
            author = author.rstrip().strip()
            cnt[author] += 1
            cnt_tot += 1
            
            
    for index, row in db['T'].iterrows():
        if type(row['author']) != str: 
            logger.error(f"Missing author in tab 'T', rid '{row['rid']}'.")
            return 1
        for author in row['author'].split(';'):
            author = author.rstrip().strip()
            cnt[author] += 1
            cnt_tot += 1
        
        
    # compute percentages:
    pct = {author: cnt[author]/cnt_tot*100 for author in cnt.keys()}
    # sort in descending order: 
    pct = dict(sorted(pct.items(), key=lambda item: item[1], reverse=True))
    # convert to string:
    pct = {author: f'{round(pct[author],2)}%' for author in pct.keys()}
    logger.debug(f"Contributions: {pct}.")
    
    
    return 0


import warnings
from decimal import Decimal


import cobra



def get_uni_biom_coeff():
    
    return 0.00000001



def check_exp_biomass_data(logger, dbexp, biomass):
    
    
    # check if the strain ID is present:
    for sheet in ['MWF', 'DNA', 'RNA', 'PROTS', 'LIPIDS_PL', 'LIPIDS_FA']:
        if biomass not in dbexp[sheet]:
            logger.error(f"ID '{biomass}' not found in sheet '{sheet}'.")
            return 1

    
    # format data
    ftd = dict()  # fraction_to_decimals
    for sheet in ['MWF', 'DNA', 'RNA', 'PROTS', 'LIPIDS_PL', 'LIPIDS_FA']:
        fraction_db = dbexp[sheet][biomass].dropna() # they should be str
        fraction_db = fraction_db.iloc[4:] # ignore ['description', 'doi', 'author', 'units']) 
        ftd[sheet] = {i: Decimal(val) for i, val in fraction_db.items() if Decimal(val) != Decimal('0.0')} # convert to dict
        if sum(ftd[sheet].values()) != Decimal('1.0'):  # check if the sum gives 1 (g/gDW or mol/mol depending on 'sheet')
            logger.error(f"Biomass data provided in sheet '{sheet}' for ID '{biomass}' does not sum up to 1.0. Missing mass is {Decimal('1.0')-sum(ftd[sheet].values())}.")
            return 1
        
    
    # normalize if ashes are present: 
    if 'ASHES' in ftd['MWF'].keys():
        logger.info(f"Found 'ASHES' > 0 in '{biomass}' data: normalizing...")
        #
        # compute sum(gDW) without ashes: 
        del ftd['MWF']['ASHES']
        new_sum = Decimal(0.0)
        for key, value in ftd['MWF'].items():
            new_sum = new_sum + value
        #
        # normalize coefficients:
        for key, value in ftd['MWF'].items():
            ftd['MWF'][key] = ftd['MWF'][key] / new_sum
    
    
    return (ftd['MWF'], ftd['DNA'], ftd['RNA'], ftd['PROTS'], ftd['LIPIDS_PL'], ftd['LIPIDS_FA'])



def rstring_builder(reactant_to_coeffs, product_to_coeffs):
    
    
    rstring = ''
    
    # parse reactants
    for rid, coeff in reactant_to_coeffs.items():
        if rstring != '':
            rstring = rstring + ' + '
        rstring = rstring + f"{coeff} {rid}"
        
    # add arrow:
    rstring = rstring + ' --> '
    
    # parse products: 
    for rid, coeff in product_to_coeffs.items():
        if rstring.endswith(' --> ') == False:
            rstring = rstring + ' + '
        rstring = rstring + f"{coeff} {rid}"
    
    
    return rstring
    
    
    
def get_fraction_dict_charge_molweight(model, reactant_to_coeffs, product_to_coeffs):
    # written for DNA, but it's general!
    
    
    # define molecular formula and charge of DNA
    DNA_dict = dict()              # for 1 mol
    DNA_charge = Decimal('0.0')    # for 1 mol
    DNA_molweight = Decimal('0.0') # for 1 mol
    #
    # reactants
    for mid, coeff in reactant_to_coeffs.items():
        m = model.metabolites.get_by_id(mid)
        for key, value in m.elements.items():
            if key not in DNA_dict.keys():
                DNA_dict[key] = Decimal('0.0')
            DNA_dict[key] = DNA_dict[key] + (Decimal(value) * coeff)        # addition
        DNA_charge = DNA_charge + (m.charge * coeff)                        # addition
        DNA_molweight = DNA_molweight + (Decimal(m.formula_weight) * coeff) # addition
    #
    # products
    for mid, coeff in product_to_coeffs.items():
        m = model.metabolites.get_by_id(mid)
        for key, value in m.elements.items():
            if key not in DNA_dict.keys():
                DNA_dict[key] = Decimal('0.0')
            DNA_dict[key] = DNA_dict[key] - (Decimal(value) * coeff)        # subtraction
        DNA_charge = DNA_charge - (m.charge * coeff)                        # subtraction
        DNA_molweight = DNA_molweight - (Decimal(m.formula_weight) * coeff) # subtraction
        
        
    return DNA_dict, DNA_charge, DNA_molweight



def compute_exp_DNA_coeffs(logger, model, MWF, DNA):
    
    
    # define ATP costs for DNA synthesis per monomer (taken from https://doi.org/10.1186/1471-2180-5-39 )
    unwinding_helix = Decimal('1.0')   # mol(ATP) / mol(DNA)
    proofreading = Decimal('0.36')   # mol(ATP) / mol(DNA)
    discontinuous_synthesis = Decimal('0.006')   # mol(ATP) / mol(DNA)
    negative_supercoiling = Decimal('0.005')   # mol(ATP) / mol(DNA)
    methylation = Decimal('0.001')   # mol(ATP) / mol(DNA)
    # real DNA synthesis takes deoxyribonucleotide-triphosphates as precursors, so DNA assembly (modeled as monophosphates) costs an additional 2 mol ATP per mol DNA.
    incorporation = Decimal('2.0')
    atp_cost_DNA = sum([unwinding_helix, proofreading, discontinuous_synthesis, negative_supercoiling, methylation, incorporation])
    
    
    # define the rstring:
    reactant_to_coeffs = {
        'damp_c': DNA['damp'],
        'dtmp_c': DNA['dtmp'],
        'dcmp_c': DNA['dcmp'],
        'dgmp_c': DNA['dgmp'],
        'atp_c': atp_cost_DNA,
        'h2o_c': atp_cost_DNA,
    }
    product_to_coeffs = {
        #'DNA_c': Decimal('1.0'),
        'adp_c': atp_cost_DNA,
        'pi_c': atp_cost_DNA,
        'h_c': atp_cost_DNA,
    }
    
    
    # define molecular formula and charge of DNA_c
    DNA_dict, DNA_charge, DNA_molweight = get_fraction_dict_charge_molweight(model, reactant_to_coeffs, product_to_coeffs)
        
        
    # adjust coefficients to match MWF
    product_to_coeffs['DNA_c'] = Decimal('1.0')
    factor = MWF['DNA'] / DNA_molweight * 1000   # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
    for mid, coeff in reactant_to_coeffs.items():
        reactant_to_coeffs[mid] = reactant_to_coeffs[mid] * factor
    for mid, coeff in product_to_coeffs.items():
        product_to_coeffs[mid] = product_to_coeffs[mid] * factor

        
    return (reactant_to_coeffs, product_to_coeffs)



def compute_exp_RNA_coeffs(logger, model, MWF, RNA):
    
    
    # define ATP costs for RNA synthesis per monomer (taken from https://doi.org/10.1186/1471-2180-5-39 )
    discarding_segments = Decimal('0.38')
    modification = Decimal('0.02')
    # real RNA synthesis takes ribonucleotide-triphosphates as precursors, so RNA assembly (modeled as monophosphates) costs an additional 2 mol ATP per mol RNA.
    incorporation = Decimal('2.0')
    atp_cost_RNA = sum([discarding_segments, modification, incorporation])
    
    
    # define the rstring:
    reactant_to_coeffs = {
        'amp_c': RNA['amp'],
        'ump_c': RNA['ump'],
        'cmp_c': RNA['cmp'],
        'gmp_c': RNA['gmp'],
        'atp_c': atp_cost_RNA,
        'h2o_c': atp_cost_RNA,
    }
    product_to_coeffs = {
        #'RNA_c': Decimal('1.0'),
        'adp_c': atp_cost_RNA,
        'pi_c': atp_cost_RNA,
        'h_c': atp_cost_RNA,
    }
    
    
    # define molecular formula and charge of RNA_c
    RNA_dict, RNA_charge, RNA_molweight = get_fraction_dict_charge_molweight(model, reactant_to_coeffs, product_to_coeffs)
        
    
    # adjust coefficients to match MWF
    product_to_coeffs['RNA_c'] = Decimal('1.0')
    factor = MWF['RNA'] / RNA_molweight * 1000   # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
    for mid, coeff in reactant_to_coeffs.items():
        reactant_to_coeffs[mid] = reactant_to_coeffs[mid] * factor
    for mid, coeff in product_to_coeffs.items():
        product_to_coeffs[mid] = product_to_coeffs[mid] * factor

        
    return (reactant_to_coeffs, product_to_coeffs)



def compute_exp_PROTS_coeffs(logger, model, MWF, PROTS):
    
    
    # define ATP costs for PROTS synthesis (taken from https://doi.org/10.1186/1471-2180-5-39 )
    activation_and_incorporation = Decimal('4.0')
    mRNA_synthesis = Decimal('0.2')
    proofreading = Decimal('0.1')
    assembly_and_modification = Decimal('0.006')
    atp_cost_PROTS = sum([activation_and_incorporation, mRNA_synthesis, proofreading, assembly_and_modification])
    
    
    # define the rstring
    aas = ['ala__L', 'arg__L', 'asn__L', 'asp__L', 'cys__L', 'gln__L', 'glu__L', 'gly', 'his__L', 'ile__L', 'leu__L', 'lys__L', 'met__L', 'phe__L', 'pro__L', 'ser__L', 'thr__L', 'trp__L', 'tyr__L', 'val__L']
    reactant_to_coeffs = {}
    for aa in aas: reactant_to_coeffs[f"{aa}_c"] = PROTS[aa]
    reactant_to_coeffs['atp_c'] = atp_cost_PROTS
    reactant_to_coeffs['h2o_c'] = atp_cost_PROTS
    product_to_coeffs = {
        #'PROTS_c': Decimal('1.0'),
        'adp_c': atp_cost_PROTS,
        'pi_c': atp_cost_PROTS,
        'h_c': atp_cost_PROTS,
    }


    # define molecular formula and charge of PROTS_c
    PROTS_dict, PROTS_charge, PROTS_molweight = get_fraction_dict_charge_molweight(model, reactant_to_coeffs, product_to_coeffs)
        
    
    # adjust coefficients to match MWF
    product_to_coeffs['PROTS_c'] = Decimal('1.0')
    factor = MWF['PROTS'] / PROTS_molweight * 1000   # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
    for mid, coeff in reactant_to_coeffs.items():
        reactant_to_coeffs[mid] = reactant_to_coeffs[mid] * factor
    for mid, coeff in product_to_coeffs.items():
        product_to_coeffs[mid] = product_to_coeffs[mid] * factor

        
    return (reactant_to_coeffs, product_to_coeffs)



def compute_exp_LIPIDS_coeffs(logger, model, MWF, LIPIDS_PL, LIPIDS_FA):
    
    
    # produce mean fatty-acid assembly reaction rstring:
    reactant_to_coeffs = {}
    product_to_coeffs = {}
    for puremid, coeff in LIPIDS_FA.items():
        reactant_to_coeffs[f"{puremid}_c"] = coeff
    product_to_coeffs[f"meanfa_c"] = 1
    rstring = rstring_builder(reactant_to_coeffs, product_to_coeffs)
    
    
    # add mean fatty-acid assembly reaction
    r = cobra.Reaction('MEANFA')
    model.add_reactions([r])
    r = model.reactions.get_by_id('MEANFA')
    r.name = "Mean fatty acid assembl reaction"
    r.build_reaction_from_string(rstring)
    r.bounds = (0, 1000)
    r.gene_reaction_rule = 'spontaneous'   
    r.update_genes_from_gpr()

    
    # determine 'L' formula and charge (charge should be -1 like every fatty acid)
    L_dict = dict()              # for 1 mol
    L_charge = Decimal('0.0')    # for 1 mol
    for puremid, coeff in LIPIDS_FA.items():
        m = model.metabolites.get_by_id(f"{puremid}_c")
        for key, value in m.elements.items():
            if key not in L_dict.keys():
                L_dict[key] = Decimal('0.0')
            L_dict[key] = L_dict[key] + (value * coeff)   # addition
        L_charge = L_charge + (m.charge * coeff)          # addition
    if model.metabolites.get_by_id(f"meanfa_c").charge != L_charge:
        logger.error("Charge of 'meanfa_c' should be -1. Please contact the developer.")
    
    
    # update formula of every metabolite having 'L' element
    for m in model.metabolites:
        m_dict = dict()
        to_update = False
        for key, value in m.elements.items():                
            if key != 'L':
                if key not in m_dict.keys():
                    m_dict[key] = Decimal('0.0')
                m_dict[key] = m_dict[key] + Decimal(value)
            else:
                to_update = True
                for key_L, value_L in L_dict.items():
                    if key_L not in m_dict.keys():
                        m_dict[key_L] = Decimal('0.0')
                    m_dict[key_L] = m_dict[key_L] + (value_L * Decimal(value))
        if to_update:
            m.formula = ''.join([f'{key}{float(value)}' for key, value in m_dict.items()])
            
            
    # define ATP costs for LIPIDS synthesis
    atp_cost_LIPIDS = Decimal('0.0')
    
    
    # define the rstring:
    reactant_to_coeffs = dict()
    for puremid, coeff in LIPIDS_PL.items():
        reactant_to_coeffs[f"{puremid}_c"] = coeff
    reactant_to_coeffs['atp_c'] = atp_cost_LIPIDS
    reactant_to_coeffs['h2o_c'] = atp_cost_LIPIDS
    product_to_coeffs = {
        #'LIPIDS_c': Decimal('1.0'),
        'adp_c': atp_cost_LIPIDS,
        'pi_c': atp_cost_LIPIDS,
        'h_c': atp_cost_LIPIDS,
    }
    
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # avoids warnings like "UserWarning: 9.823 is not an integer (in formula C9.823H5.5789999999999935N3.6769999999999996O3.628P1.0)"
    
        # define molecular formula and charge of LIPIDS
        LIPIDS_dict, LIPIDS_charge, LIPIDS_molweight = get_fraction_dict_charge_molweight(model, reactant_to_coeffs, product_to_coeffs)
     
        
    # adjust coefficients to match MWF
    product_to_coeffs['LIPIDS_c'] = Decimal('1.0')
    factor = MWF['LIPIDS'] / LIPIDS_molweight * 1000   # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
    for mid, coeff in reactant_to_coeffs.items():
        reactant_to_coeffs[mid] = reactant_to_coeffs[mid] * factor
    for mid, coeff in product_to_coeffs.items():
        product_to_coeffs[mid] = product_to_coeffs[mid] * factor
    
    
    return (reactant_to_coeffs, product_to_coeffs)
    


def compute_exp_WTA_coeffs(logger, model, biomass_dict, MWF, dbexp):
    product_to_coeffs = dict()  # empty dict (no product)
    
    
    # get WTA species in biomass: 
    available_WTA = set()
    for mid in biomass_dict['WTA']: 
        if mid in [m.id for m in model.reactions.get_by_id('Biomass').reactants]:
            available_WTA.add(mid)
            
            
    # if no pathway was completed for LTA, but a fraction was given in MWF:
    if available_WTA == set() and 'WTA' in MWF.keys():
        logger.error(f"Strain seems unable to synthetize WTA, but a fraction was indicated in 'MWF'.")
        return 1
            
    
    # if no grams were specified in MWF:
    if available_WTA != set() and 'WTA' not in MWF.keys():
        logger.info(f"This strain should be able to synthetize WTA ({available_WTA}), but no fraction was indicated in 'MWF': they will be treated as cofactors (coefficient {get_uni_biom_coeff()}).")
        reactant_to_coeffs = dict()
        for mid in available_WTA:
            reactant_to_coeffs[mid] = Decimal(get_uni_biom_coeff())
        return (reactant_to_coeffs, product_to_coeffs)
    
    
    # equal grams are assumed for every detected species of WTA
    reactant_to_coeffs = dict()
    for mid in available_WTA:
        m = model.metabolites.get_by_id(mid) 
        # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
        reactant_to_coeffs[mid] = MWF['WTA'] / len(available_WTA) / Decimal(m.formula_weight) * 1000

        
    return (reactant_to_coeffs, product_to_coeffs)



def compute_exp_LTA_coeffs(logger, model, biomass_dict, MWF, dbexp):
    product_to_coeffs = dict()  # empty dict (no product)
    
    
    # get LTA species in biomass: 
    available_LTA = set()
    for mid in biomass_dict['LTA']: 
        if mid in [m.id for m in model.reactions.get_by_id('Biomass').reactants]:
            available_LTA.add(mid)
            
            
    # if no pathway was completed for LTA, but a fraction was given in MWF:
    if available_LTA == set() and 'LTA' in MWF.keys():
        logger.error(f"Strain seems unable to synthetize LTA, but a fraction was indicated in 'MWF'.")
        return 1
            
    
    # if no grams were specified in MWF:
    if available_LTA != set() and 'LTA' not in MWF.keys():
        logger.info(f"This strain should be able to synthetize LTA ({available_LTA}), but no fraction was indicated in 'MWF': they will be treated as cofactors (coefficient {get_uni_biom_coeff()}).")
        reactant_to_coeffs = dict()
        for mid in available_LTA:
            reactant_to_coeffs[mid] = Decimal(get_uni_biom_coeff())
        return (reactant_to_coeffs, product_to_coeffs)
    
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # avoids warnings like "UserWarning: 9.823 is not an integer (in formula C9.823H5.5789999999999935N3.6769999999999996O3.628P1.0)"
    

        # equal grams are assumed for every detected species of LTA
        reactant_to_coeffs = dict()
        for mid in available_LTA:
            m = model.metabolites.get_by_id(mid)
            # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
            reactant_to_coeffs[mid] = MWF['LTA'] / len(available_LTA) / Decimal(m.formula_weight) * 1000

        
    return (reactant_to_coeffs, product_to_coeffs)



def compute_exp_PEPTIDO_coeffs(logger, model, biomass_dict, MWF, dbexp):
    product_to_coeffs = dict()  # empty dict (no product)
    
    
    # get WTA species in biomass: 
    available_PEPTIDO = set()
    for mid in biomass_dict['PEPTIDO']: 
        if mid in [m.id for m in model.reactions.get_by_id('Biomass').reactants]:
            available_PEPTIDO.add(mid)
            
            
    # if no pathway was completed for LTA, but a fraction was given in MWF:
    if available_PEPTIDO == set() and 'PEPTIDO' in MWF.keys():
        logger.error(f"Strain seems unable to synthetize PEPTIDO, but a fraction was indicated in 'MWF'.")
        return 1
            
    
    # if no grams were specified in MWF:
    if available_PEPTIDO != set() and 'PEPTIDO' not in MWF.keys():
        logger.info(f"This strain should be able to synthetize PEPTIDO ({available_PEPTIDO}), but no fraction was indicated in 'MWF': they will be treated as cofactors (coefficient {get_uni_biom_coeff()}).")
        reactant_to_coeffs = dict()
        for mid in available_PEPTIDO:
            reactant_to_coeffs[mid] = Decimal(get_uni_biom_coeff())
        return (reactant_to_coeffs, product_to_coeffs)
    
    
    # equal grams are assumed for every detected species of WTA
    reactant_to_coeffs = dict()
    for mid in available_PEPTIDO:
        m = model.metabolites.get_by_id(mid) 
        # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
        reactant_to_coeffs[mid] = MWF['PEPTIDO'] / len(available_PEPTIDO) / Decimal(m.formula_weight) * 1000

        
    return (reactant_to_coeffs, product_to_coeffs)



def compute_exp_EPS_coeffs(logger, model, biomass_dict, MWF, dbexp):
    product_to_coeffs = dict()  # empty dict (no product)
    
    
    # get WTA species in biomass: 
    available_EPS = set()
    for mid in biomass_dict['EPS']: 
        if mid in [m.id for m in model.reactions.get_by_id('Biomass').reactants]:
            available_EPS.add(mid)
            
            
    # if no pathway was completed for LTA, but a fraction was given in MWF:
    if available_EPS == set() and 'EPS' in MWF.keys():
        logger.error(f"Strain seems unable to synthetize EPS, but a fraction was indicated in 'MWF'.")
        return 1
            
    
    # if no grams were specified in MWF:
    if available_EPS != set() and 'EPS' not in MWF.keys():
        logger.info(f"This strain should be able to synthetize EPS ({available_EPS}), but no fraction was indicated in 'MWF': they will be treated as cofactors (coefficient {get_uni_biom_coeff()}).")
        reactant_to_coeffs = dict()
        for mid in available_EPS:
            reactant_to_coeffs[mid] = Decimal(get_uni_biom_coeff())
        return (reactant_to_coeffs, product_to_coeffs)
    
    
    # equal grams are assumed for every detected species of WTA
    reactant_to_coeffs = dict()
    for mid in available_EPS:
        m = model.metabolites.get_by_id(mid) 
        # [g/gDW] / [g/mol] * 1000 = [mmol/gDW]
        reactant_to_coeffs[mid] = MWF['EPS'] / len(available_EPS) / Decimal(m.formula_weight) * 1000

        
    return (reactant_to_coeffs, product_to_coeffs)




import pandas as pnd
import gempipe



from ..commons import get_biomass_dict
from ..commons import apply_medium_given_column
from ..commons import get_biomass_df_structure



def precursors_on_media(logger, model, universe, dbexp, media, cond_col_dict, precursors):
    
                
    # check if requested
    df_E = pnd.DataFrame()
    if not precursors:
        return df_E
    df_E = get_biomass_df_structure(logger, df_E, model, universe, dbexp, cond_col_dict)

        
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: blocked biomass precursors check will be skipped.")
        return df_E
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in dbexp['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(dbexp['media'].columns)}.")
        return 1
            
    
    logger.info(f"Checking blocked biomass precursors on {len(media)} media...")
    for medium in media:
        if medium not in dbexp['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_E[f'{medium}'] = None
        logger.debug(f"Checking blocked biomass precursors on medium '{medium}'...")
            
            
        # apply medium :
        response = apply_medium_given_column(logger, model, medium, dbexp['media'][medium])
        if response == 1: return 1
    
            
        # check separately for each biomass precursor:
        # force 'atp_c' as first:
        universal_precs = ['atp_c'] + [m.id for m in universe.reactions.Biomass.reactants if m.id not in ['atp_c', 'h2o_c']]
        for mid in universal_precs:
            # skip conditional precursors: 
            if mid in [m.id for m in model.reactions.Biomass.reactants]:
                
            
                response, flux, status = gempipe.can_synth(model, mid=mid)
                if response == True:
                    df_E.loc[mid, f'{medium}'] = '/'
                else:
                    df_E.loc[mid, f'{medium}'] = 'blocked'
    
    
    return df_E



import importlib.metadata 


import pandas as pnd


from .figures import figure_df_C_F1




def get_summary_sheet(model, memote_results_dict):
    df_gsrap = [
        # Gsrap
        {'c1': 'Gsrap version', 'c2': f"v{importlib.metadata.metadata('gsrap')['Version']}", 'c3': '', 'c4': ''},
        {'c1': 'Model ID', 'c2': f"{model.id}", 'c3': '', 'c4': ''},
        {'c1': 'Compartments', 'c2': f"{len(model.compartments)}", 'c3': '', 'c4': ''},
        {'c1': 'Metabolites', 'c2': f"{len(model.metabolites)}", 'c3': '', 'c4': ''},
        {'c1': '', 'c2': 'Unique', 'c3': f"{len(set([m.id.rsplit('_',1)[0] for m in model.metabolites]))}", 'c4': ''},
        {'c1': 'Reactions', 'c2': f"{len(model.reactions)}", 'c3': '', 'c4': ''},
        {'c1': '', 'c2': 'Non-transport', 'c3': f"{len([r for r in model.reactions if ((r.id != 'Biomass' and len(r.metabolites)!=1) and len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))==1)])}", 'c4': ''},
        {'c1': '', 'c2': 'Transport', 'c3': f"{len([r for r in model.reactions if ((r.id != 'Biomass' and len(r.metabolites)!=1) and len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))>1)])}", 'c4': ''},
        {'c1': '', 'c2': 'Artificial', 'c3': f"{len([r for r in model.reactions if ((r.id == 'Biomass' or len(r.metabolites)==1))])}", 'c4': ''},
        {'c1': 'Genes', 'c2': f"{len(model.genes)}", 'c3': '', 'c4': ''},
        # empty line
        {'c1': '', 'c2': '', 'c3': '', 'c4': ''},
    ]
    df_gsrap = pnd.DataFrame.from_records(df_gsrap)
    if memote_results_dict != None:
        df_memote = [
            # MEMOTE
            {'c1': 'MEMOTE version', 'c2': f"v{memote_results_dict['version']}", 'c3': '', 'c4': ''},
            {'c1': 'MEMOTE Total Score', 'c2': f"{memote_results_dict['total']}%", 'c3': '', 'c4': ''},
            {'c1': 'MEMOTE consistency', 'c2': f"{memote_results_dict['consistency']['sub_total']}%", 'c3': '', 'c4': ''},
            {'c1': '', 'c2': 'stoichiometric consistency', 'c3': f"{memote_results_dict['consistency']['test_stoichiometric_consistency']}%", 'c4': ''},
            {'c1': '', 'c2': 'mass balance', 'c3': f"{memote_results_dict['consistency']['test_reaction_mass_balance']}%", 'c4': ''},
            {'c1': '', 'c2': 'charge balance', 'c3': f"{memote_results_dict['consistency']['test_reaction_charge_balance']}%", 'c4': ''},
            {'c1': '', 'c2': 'disconnected metabolites', 'c3': f"{memote_results_dict['consistency']['test_reaction_charge_balance']}%", 'c4': ''},
            {'c1': '', 'c2': 'unbounded flux in default conditions', 'c3': f"{memote_results_dict['consistency']['test_find_reactions_unbounded_flux_default_condition']}%", 'c4': ''},
            {'c1': 'MEMOTE annotation Metabolites', 'c2': f"{memote_results_dict['annotation_M']['sub_total']}%", 'c3': '', 'c4': ''},
            {'c1': 'MEMOTE annotation Reactions', 'c2': f"{memote_results_dict['annotation_R']['sub_total']}%", 'c3': '', 'c4': ''},
            {'c1': 'MEMOTE annotation Genes', 'c2': f"{memote_results_dict['annotation_G']['sub_total']}%", 'c3': '', 'c4': ''},
            {'c1': 'MEMOTE annotation SBO', 'c2': f"{memote_results_dict['annotation_SBO']['sub_total']}%", 'c3': '', 'c4': ''},
        ]
        df_memote = pnd.DataFrame.from_records(df_memote)
    else:
        df_memote = pnd.DataFrame()

        
    df = pnd.concat([df_gsrap, df_memote])
    return df



def write_excel_model(model, filepath, nofigs, memote_results_dict, df_E, df_B, df_P, df_S, df_C=None):
    
    
    # generate figures
    if nofigs == False:
              
        if df_C is not None:
            df_C_F1 = figure_df_C_F1(df_C)
        
        
    
    # format df_E:  # biomass precursors biosynthesis
    if df_E is not None:
        df_E.insert(0, 'mid', '')  # new columns as first
        df_E['mid'] = df_E.index
        df_E = df_E.reset_index(drop=True)
    
    # format df_B:  # biomass assembly
    if df_B is not None:
        df_B.insert(0, 'mid', '')  # new columns as first
        df_B['mid'] = df_B.index
        df_B = df_B.reset_index(drop=True)
    
    # format df_P:  phenotype screening (Biolog(R))
    if df_P is not None:
        df_P.insert(0, 'plate:well', '')  # new columns as first
        df_P['plate:well'] = df_P.index
        df_P = df_P.reset_index(drop=True)
        
    # format df_S:  metabolite synthesis
    if df_S is not None:
        df_S.insert(0, 'mid', '')  # new columns as first
        df_S['mid'] = df_S.index
        df_S = df_S.reset_index(drop=True)
        
    # format df_C: universal reaction coverage
    if df_C is not None:
        df_C.insert(0, 'kr', '')  # new columns as first
        df_C['kr'] = df_C.index
        df_C = df_C.reset_index(drop=True)
            
        
    
    # define dict-lists, future dataframes
    df_M = []
    df_R = []
    df_T = []
    df_G = []
    df_A = []
    
    for m in model.metabolites: 
        row_dict = {'mid': m.id, 'name': m.name, 'formula': m.formula, 'charge': m.charge,}
        
        for db in m.annotation.keys():
            annots = m.annotation[db]
            if type(annots) == str: annots = [annots]
            annots = '; '.join([i for i in annots])
            row_dict[db] = annots
        df_M.append(row_dict)
        
    for r in model.reactions:
        row_dict = {'rid': r.id, 'name': r.name, 'rstring': r.reaction, 'gpr': "Not applicable", 'bounds': r.bounds}
        
        for db in r.annotation.keys():
            annots = r.annotation[db]
            if type(annots) == str: annots = [annots]
            annots = '; '.join([i for i in annots])
            row_dict[db] = annots
        
        # handle artificial reactions
        if r.id == 'Biomass':
            # commented as the type is inplicit in the ID
            #row_dict['type'] = 'biomass'
            df_A.append(row_dict)
                
        elif len(r.metabolites) == 1:
            # commented as the type is inplicit in the ID
            """
            if len(r.metabolites)==1 and list(r.metabolites)[0].id.rsplit('_',1)[-1] == 'e': 
                row_dict['type'] = 'exchange'
            elif r.lower_bound < 0 and r.upper_bound > 0:
                row_dict['type'] = 'sink'
            elif r.lower_bound == 0 and r.upper_bound > 0:
                row_dict['type'] = 'demand'
            """
            df_A.append(row_dict)
        
        else: # more than 1 metabolite involved
            row_dict['gpr'] = r.gene_reaction_rule
            
            # introduce reaction in the correct table: 
            if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites])) == 1:
                df_R.append(row_dict)
            else: df_T.append(row_dict)
            
    for g in model.genes:
        row_dict = {'gid': g.id, 'name': g.name, 'involved_in': '; '.join([r.id for r in g.reactions])}
        
        for db in g.annotation.keys():
            annots = g.annotation[db]
            if type(annots) == str: annots = [annots]
            annots = '; '.join([i for i in annots])
            row_dict[db] = annots
        df_G.append(row_dict)
    
    # create dataframes from dict-lists
    df_M = pnd.DataFrame.from_records(df_M)
    df_R = pnd.DataFrame.from_records(df_R)
    df_T = pnd.DataFrame.from_records(df_T)
    df_A = pnd.DataFrame.from_records(df_A)
    df_G = pnd.DataFrame.from_records(df_G)
    
    # sort columns
    df_M_first_cols = ['mid', 'name', 'formula', 'charge']
    df_M = df_M[df_M_first_cols + sorted([c for c in df_M.columns if c not in df_M_first_cols])]
    df_R_first_cols = ['rid', 'name', 'rstring', 'gpr', 'bounds']
    df_R = df_R[df_R_first_cols + sorted([c for c in df_R.columns if c not in df_R_first_cols])]
    df_T = df_T[df_R_first_cols + sorted([c for c in df_T.columns if c not in df_R_first_cols])]
    df_A = df_A[df_R_first_cols + sorted([c for c in df_A.columns if c not in df_R_first_cols])]
    df_G_first_cols = ['gid', 'name', 'involved_in']
    df_G = df_G[df_G_first_cols + sorted([c for c in df_G.columns if c not in df_G_first_cols])]
    
    
    
    with pnd.ExcelWriter(filepath, engine='xlsxwriter') as writer:
        get_summary_sheet(model, memote_results_dict).to_excel(writer, sheet_name='Summary', index=False, header=False)
        df_M.to_excel(writer, sheet_name='Metabolites', index=False)
        df_R.to_excel(writer, sheet_name='Reactions', index=False)
        df_T.to_excel(writer, sheet_name='Transporters', index=False)
        df_G.to_excel(writer, sheet_name='Genes', index=False)
        df_A.to_excel(writer, sheet_name='Artificials', index=False)
        if df_E is not None and len(df_E)!=0: df_E.to_excel(writer, sheet_name='Precursors', index=False)
        if df_B is not None: df_B.to_excel(writer, sheet_name='Biomass', index=False)
        if df_P is not None and len(df_P)!=0: df_P.to_excel(writer, sheet_name='Biolog®', index=False)
        if df_S is not None and len(df_S.columns)>2: df_S.to_excel(writer, sheet_name='Biosynth', index=False) 
        if df_C is not None: 
            df_C.to_excel(writer, sheet_name='Coverage', index=False) 
            if nofigs == False:
                worksheet = writer.sheets['Coverage']
                worksheet.insert_image('E3', 'df_C_F1.png', {'image_data': df_C_F1})
            
        
    sheets_dict = {
        'model_id': model.id,
        'Metabolites': df_M,
        'Reactions': df_R,
        'Transporters': df_T,
        'Artificials': df_A,
        'Precursors': df_E,
        'Biomass': df_B,
        'Biolog': df_P,
        'Biosynth': df_S,
        'Coverage': df_C,
    }
    return sheets_dict



def comparative_table(logger, outdir, sheets_dicts):
    
    
    # create topology table (reaction presence / absence matrix): 
    df_topology = pnd.DataFrame()
    for sheets_dict in sheets_dicts:
        for index, row in sheets_dict['Reactions'].iterrows():
            if row['rid'] not in df_topology.index:
                df_topology.loc[row['rid'], 'rid'] = row['rid']
                for key, value in row.to_dict().items():
                    # force string to avoid errors with bounds
                    df_topology.loc[row['rid'], key] = '' if pnd.isna(value) else str(value)   
            df_topology.loc[row['rid'], sheets_dict['model_id']] = 1
    for sheets_dict in sheets_dicts:  # replace missing values:
        df_topology = df_topology.fillna({sheets_dict['model_id']: 0})
        
        
    # create GPR table (reaction presence / absence matrix): 
    df_gprs = pnd.DataFrame()
    for sheets_dict in sheets_dicts:
        for index, row in sheets_dict['Reactions'].iterrows():
            if row['rid'] not in df_gprs.index:
                df_gprs.loc[row['rid'], 'rid'] = row['rid']
                for key, value in row.to_dict().items():
                    # force string to avoid errors with bounds
                    df_gprs.loc[row['rid'], key] = '' if pnd.isna(value) else str(value)   
            df_gprs.loc[row['rid'], sheets_dict['model_id']] = row['gpr']
    for sheets_dict in sheets_dicts:  # replace missing values:
        df_gprs = df_gprs.fillna({sheets_dict['model_id']: 'missing'})
        
    
    with pnd.ExcelWriter(f"{outdir}/comparison.mkmodel.xlsx") as writer:
        df_topology.to_excel(writer, sheet_name='Topology', index=False)
        df_gprs.to_excel(writer, sheet_name='GPRs', index=False)
    logger.info(f"'{outdir}/comparison.mkmodel.xlsx' created!")
        
    


import logging
import sys

import colorlog



def set_header_trailer_formatter(handler):
    # handy function to print without time/level (for header / trailer)
    
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)

    

def set_usual_formatter(handler):
    # to print the main pipeline logging:
    
    formatter = colorlog.ColoredFormatter(
        "%(asctime)s %(log_color)s%(levelname)s%(reset)s: %(message)s", datefmt="%H:%M:%S",
        log_colors={'DEBUG': 'cyan', 'INFO': 'green', 'WARNING': 'yellow', 'ERROR': 'red', 'CRITICAL': 'bold_red'})
    handler.setFormatter(formatter)


    
def get_logger(logger_id, verbose):

    
    # create a clean logger without handlers
    logger = logging.getLogger(logger_id)    
    if logger.hasHandlers(): 
        logger.handlers.clear()
        
    # stop bubbling up to root logger
    logger.propagate = False  
    
    
    # show everything on the stderr:
    handler = logging.StreamHandler(stream=sys.stderr)
    logger.addHandler(handler)
    
    # set the level 
    if verbose: logger.setLevel(logging.DEBUG) # debug (lvl 10) and up
    else:       logger.setLevel(logging.INFO) # debug (lvl 20) and up
    
    # attach the default formatter
    set_usual_formatter(handler)
    
    
    return logger
    
    



import os
import pickle
from importlib import resources


import cobra
import gempipe


from ..commons import force_id_on_sbml
from ..commons import log_metrics
from ..commons import log_unbalances
from ..commons import get_databases
from ..commons import introduce_universal_biomass
from ..commons import write_excel_model
from ..commons import show_contributions
from ..commons import adjust_biomass_precursors
from ..commons import count_undrawn_rids
from ..commons import format_expansion
from ..commons import check_taxon
from ..commons import download_keggorg
from ..commons import initialize_model
from ..commons import get_memote_results_dict


from .introduce import introduce_metabolites
from .introduce import introduce_reactions
from .introduce import introduce_transporters
from .introduce import introduce_sinks_demands

from .annotation import translate_annotate_genes
from .annotation import set_up_groups

from .completeness import check_completeness

from ..runsims.simplegrowth import grow_on_media
from ..runsims.precursors import precursors_on_media
from ..runsims.biosynth import biosynthesis_on_media

from ..mkmodel.polishing import remove_disconnected

from .cycles import verify_egc_all




def main(args, logger):
    
    
    ###### FORMAT ARGS NOT REQUIRING RESOURCES
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # check compatibility of input parameters
    if args.progress==False and args.module==True: 
        logger.error(f"You cannot ask --module without --progress (see --help).")
        return 1
    if args.progress==False and args.focus!='-':
        logger.error(f"You cannot ask --focus without --progress (see --help).")
        return 1
    
    
    # check 'goodbefore' and 'onlyauthor' params
    if args.goodbefore == '-' and args.onlyauthor != '-':
        logger.error(f"--onlyauthor must be used in conjunction with --goodbefore.")
        return 1
    if   args.goodbefore == '-': args.goodbefore = [None, None, None]
    elif len(args.goodbefore.split('-')) != 3: 
        logger.error(f"Invalid syntax detected for --goodbefore.")
        return 1
    else:
        args.goodbefore = args.goodbefore.split('-')
        if args.goodbefore[0] == 'None': args.goodbefore[0] = None
        if args.goodbefore[1] == 'None': args.goodbefore[1] = None
        if args.goodbefore[2] == 'None': args.goodbefore[2] = None
    if args.onlyauthor == '-': args.onlyauthor = None
    
    
    
    ###### LOAD LOCAL RESOURCES
    # check and extract the required 'gsrap.maps' file
    if os.path.exists(f'{args.inmaps}') == False:
        logger.error(f"File 'gsrap.maps' not found at {args.inmaps}.")
        return 1
    try: 
        with open(f'{args.inmaps}', 'rb') as f:
            inmaps = pickle.load(f)  
    except: 
        logger.error(f"Provided file {args.inmaps} has an incorrect format.")
        return 1
    idcollection_dict = inmaps['idcollection_dict']
    summary_dict = inmaps['summary_dict']
    
    
    # load internal resources
    with resources.path("gsrap.assets", f"kegg_compound_to_others.pickle") as asset_path: 
        with open(asset_path, 'rb') as handle:
            kegg_compound_to_others = pickle.load(handle)
    with resources.path("gsrap.assets", f"kegg_reaction_to_others.pickle") as asset_path: 
        with open(asset_path, 'rb') as handle:
            kegg_reaction_to_others = pickle.load(handle) 
            
            
            
    ###### FORMAT/CHECK FOCUSING ARGS
    # format the --eggnog param
    args.eggnog = format_expansion(logger, args.eggnog)  # now 'args.eggnog' could still be '-'
    
    # check the --taxon param
    if args.taxon != '-':
        response = check_taxon(logger, args.taxon, idcollection_dict)
        if response == 1: return 1
    
    # get the kegg organism if requested
    if args.keggorg != '-':
        response = download_keggorg(logger, args.keggorg, args.outdir)
        if response == 1: return 1
    
    
    
    # DOWNLOAD ONLINE RESOURCES
    # get dbuni and dbexp:
    logger.info("Downloading gsrap database...")
    response = get_databases(logger)
    if type(response)==int: return 1
    else: dbuni, dbexp, lastmap = response
    

    # show simple statistics of contributions
    response = show_contributions(logger, dbuni, args.goodbefore)
    if response == 1: return 1
                                    
        
    
    ###### RECONSTRUCTION
    # create the model
    universe = cobra.Model('universe')   
    logger.info("Parsing gsrap database...")
    
    # introduce M / R / T
    universe = introduce_metabolites(logger, dbuni, universe, idcollection_dict, kegg_compound_to_others, args.goodbefore[0], args.onlyauthor)
    if type(universe)==int: return 1
    universe = introduce_reactions(logger, dbuni, universe, idcollection_dict, kegg_reaction_to_others, args.goodbefore[1], args.onlyauthor)
    if type(universe)==int: return 1
    universe = introduce_transporters(logger, dbuni, universe, idcollection_dict, kegg_reaction_to_others, args.goodbefore[2], args.onlyauthor)
    if type(universe)==int: return 1

    # introduce sinks / demands (exchanges where included during T)
    universe = introduce_sinks_demands(logger, universe)
    if type(universe)==int: return 1

    # introducce biomass
    universe = introduce_universal_biomass(logger, dbexp, universe)
    if type(universe)==int: return 1



    ###### ANNOTATION
    # translate Gs to symbols and annotate them (EC, COG, GO, ...)
    universe = translate_annotate_genes(logger, universe, idcollection_dict)
    if type(universe)==int: return 1

    # introduce collectionas (groups of Rs as maps/modules)
    universe = set_up_groups(logger, universe, idcollection_dict)
    if type(universe)==int: return 1


    
    # # # # #   PARSING ENDS HERE   # # # # #
    log_metrics(logger, universe)
    log_unbalances(logger, universe)

    
    
    ###### CHECKS 1
    # check universe completness
    df_C = check_completeness(logger, universe, args.progress, args.module, args.focus, args.taxon, args.eggnog, args.keggorg, idcollection_dict, summary_dict, args.outdir)
    if type(df_C)==int: return 1



    ###### POLISHING 1
    # remove disconnected metabolites
    universe = remove_disconnected(logger, universe)

    
    
    ###### CHECKS 2
    # check erroneous EGCs
    verify_egc_all(logger, universe, args.outdir)
    
    
    if not args.justparse:
        
        
        ###### CHECKS 3
        # check growth on minmal media
        df_G = grow_on_media(logger, universe, dbexp, args.media, '-', True)
        if type(df_G)==int: return 1

        # check blocked biomass precursors
        cond_col_dict = adjust_biomass_precursors(logger, universe, universe, 1.0)
        df_E = precursors_on_media(logger, universe, universe, dbexp, args.media, cond_col_dict, args.precursors)
        if type(df_E)==int: return 1

        # check blocked metabolites / dead-ends
        df_S = biosynthesis_on_media(logger, universe, dbexp, args.media, args.biosynth)
        if type(df_S)==int: return 1


    
        ###### POLISHING 2
        # reset growth environment befor saving the model
        gempipe.reset_growth_env(universe)

        # initialize model
        response = initialize_model(logger, universe, dbexp, args.initialize, args.media)
        if response==1: return 1



        ###### CHECKS 4
        # compute Memote metrics
        memote_results_dict = get_memote_results_dict(logger, universe)



        # output the universe
        logger.info("Writing universal model...")
        cobra.io.save_json_model(universe, f'{args.outdir}/universe.json')
        logger.info(f"'{args.outdir}/universe.json' created!")
        cobra.io.write_sbml_model(universe, f'{args.outdir}/universe.xml')   # groups are saved only to SBML 
        logger.info(f"'{args.outdir}/universe.xml' created!")
        force_id_on_sbml(f'{args.outdir}/universe.xml', 'universe')   # force introduction of the 'id=""' field
        sheets_dict = write_excel_model(universe, f'{args.outdir}/universe.parsedb.xlsx', args.nofigs, memote_results_dict, df_E, None, None, df_S, df_C)  
        logger.info(f"'{args.outdir}/universe.parsedb.xlsx' created!")


    
    ###### CHECKS 4
    # check if universal escher map is updated:
    count_undrawn_rids(logger, universe, lastmap)
    
    
    return 0




def print_json_tree(data, level=0, max_level=2):
    # explore contents of a json object
    
    if level > max_level:
        return
    indent = '  ' * level
    if isinstance(data, dict):
        for key, value in data.items():
            print(f"{indent}{key}")
            print_tree(value, level + 1, max_level)
    elif isinstance(data, list):
        for i, item in enumerate(data):
            print(f"{indent}[{i}]")
            print_tree(item, level + 1, max_level)
            
            

def count_undrawn_rids(logger, universe, lastmap):
    
    
    rids = set([r.id for r in universe.reactions])
    
    drawn_rids = set()
    for key, value in lastmap['json'][1]['reactions'].items():
        drawn_rids.add(value['bigg_id'])
        
        
    remainings = rids - drawn_rids
    filename = lastmap['filename']
    logger.debug(f"Last universal map version detected: '{filename}'.")
    if len(remainings) > 0:
        logger.warning(f"Our universal map is {len(remainings)} reactions behind. Please draw!")
    else:
        logger.info(f"Our universal map is {len(remainings)} reactions behind. Thank you ♥")
        


import warnings
from decimal import Decimal


import numpy as np
import pandas as pnd
import cobra


from .coeffs import get_uni_biom_coeff
from .coeffs import check_exp_biomass_data
from .coeffs import compute_exp_DNA_coeffs
from .coeffs import compute_exp_RNA_coeffs
from .coeffs import compute_exp_PROTS_coeffs
from .coeffs import compute_exp_LIPIDS_coeffs
from .coeffs import compute_exp_WTA_coeffs
from .coeffs import compute_exp_LTA_coeffs
from .coeffs import compute_exp_PEPTIDO_coeffs
from .coeffs import compute_exp_EPS_coeffs



def get_biomass_dict(logger, universe, dbexp):
    
    
    # check if precursors are modeled in the universe:
    universal_mids = [m.id for m in universe.metabolites]
    fraction_to_precursors = dict()
    for fraction in ['DNA', 'RNA', 'PROTS', 'LIPIDS_PL', 'LIPIDS_FA']:
        fraction_db = dbexp[fraction]
        fraction_db = fraction_db.iloc[4:] # ignore ['description', 'doi', 'author', 'units']) 
        precursors = [f"{i}_c" for i in fraction_db.index.dropna()]
        for i in precursors: 
            if i not in universal_mids:
                logger.error(f"Metabolite '{i}', indicated in 'DNA', is not present in universe.")
                return 1
        fraction_to_precursors[fraction] = precursors
    
    
    biomass_dict = {
        
        
        # MWF: macromolecular weight fractions:
        'DNA': fraction_to_precursors['DNA'],
        'RNA': fraction_to_precursors['RNA'],
        'PROTS': fraction_to_precursors['PROTS'],
        'LIPIDS_PL': fraction_to_precursors['LIPIDS_PL'],
        'LIPIDS_FA': fraction_to_precursors['LIPIDS_FA'],
        'WTA': [
            'WTAgg40r_20n_20a_P_c', # teichoic acids
            'WTArg40r_20g_20a_P_c', # teichoic acids
            'WTAg40g_20g_20a_P_c', # teichoic acids
            'WTAt4_20a_P_c', # teoichoic acid
        ],
        'LTA': [
            'LTAgg40g_20n_20a_c', # lipoteichoic acids
            'LTAga40g_20t_20a_c', # lipoteichoic acids 
            'LTAt4_20a_c', # lipoteichoic acid
        ],
        'PEPTIDO': [
            'peptidoSTA_c', # peptidoglycan (dependant on 'udcpdp_c')
            'peptidoSTR_c',
            'peptidoDAP_c',
        ],
        'EPS': [
            'epsLAB_c', # LAB eps            
        ],
        # glycogen
        # starch
        'PHA': [
            'phb_c', # PHA / PHB
        ],
        
        
        # Cofactors:
        # Note: universal ('cofs_uni_Xavier2017') and conditional ('cofs_cond_Xavier2017') cofactors have been defined in Xavier2017 .
        # Here presented in the same order of Xavier2017.
        # Cofactors not mentioned in Xavier2017 ('cofactors_add') are reported below.
        # Xavier2017: 10.1016/j.ymben.2016.12.002.
        'cofs_uni_Xavier2017': [
            'nad_c',    # B3: Nicotinamide -adenine dinucleotide phosphate
            'nadp_c',   # B3: Nicotinamide -adenine dinucleotide phosphate
            'coa_c',    # B5: Coenzyme A  (dependant on 'pnto__R_c')
            'fad_c',    # B2: Flavin adenine dinucleotide
            'fmn_c',    # B2: Flavin mononucleotide
            'ribflv_c', # B2: ribovlavin. Non-active form acording to Xavier2017 but anyway included.
            #'f4200_c', # B2: included by Xavier2017 in 'universal' but the description seems conditional.
            'thf_c',    # B9: tetrahydrofolate 
            '10fthf_c', # B9: 10-Formyltetrahydrofolate
            '5mthf_c',  # B9: 5-Methyltetrahydrofolate
            'thmpp_c',  # B1: Thiamine diphosphate
            'pydx5p_c', # B6: pyridoxal 5-phosphate
            'amet_c',   # SAM: S-adenosyl-methionine
        ],
        'cofs_cond_Xavier2017': [
            #'f4200_c', # coenzyme f420 (electron transfer in methanogens, actinobacteria , and others)
            'ptrc_c',   # Putrescine
            'spmd_c',   # Sperimidine
            'pheme_c',  # protoheme
            'mql8_c',   # menaquinol / manaquinone (mqn8_c)
            'q8h2_c',   # ubiquinol / ubiquinone (q8_c)
            # Methionaquinone
            'btn_c',    # B7: biotin
            #'ACP_c',    # Acyl-carrier protein (removed as it contains the 'X' atom, causing error in Memote)
            'adocbl_c', # B12: Adenosylcobalamin
            # Lipoate
            'uacgam_c'  # uridine diphosphate N-Acetylglucosamine (UDP-GlcNAc)
        ],
        ##### ADDED ##### (conditionals not included or lumped in Xavier2017)
        'cofs_suppl': [
            'hemeO_c',  # heme-O
            'sheme_c',  # siroheme
            'moco_c',   # molybdenum cofactor
            'phllqol_c',# phylloquinol / phylloquinone (phllqne_c)
            'gthrd_c',  # glutathione (reduced)
            'br_c',     # bacterioruberin
            # kdo_lipid_A
        ],
    }
    
    
    return biomass_dict



def get_precursors_to_pathway():
    
    
    precursors_to_pathway = {  # listing alternative biosynthetic pathways
        
        'pheme_c': ['M00868', 'M00121', 'M00926'],   # protoheme (heme)
        'mql8_c': ['M00116', 'M00930', 'M00931'],  # menaquinol
        'q8h2_c': ['M00117', 'M00989', 'M00128'],  # ubiquinol
        'adocbl_c': ['M00122'],   # vitamin B12 (cobolamin)
        'hemeO_c': ['gr_HemeO'],  # heme-O
        'sheme_c': ['M00846'],   # siroheme
        'moco_c': ['M00880'],  # molybdenum cofactor
        'phllqol_c': ['M00932'],  # phylloquinol
        'gthrd_c': ['M00118'],  # Reduced glutathione
        'br_c': ['gr_br'],  # bacterioruberin
        'WTAg40g_20g_20a_P_c':  ['gr_WTA1'], # teichoic acids
        'WTArg40r_20g_20a_P_c': ['gr_WTA2'], # teichoic acids
        'WTAgg40r_20n_20a_P_c': ['gr_WTA3'], # teichoic acids
        'WTAt4_20a_P_c': ['gr_WTA4'], # teoichoic acid
        'LTAgg40g_20n_20a_c':  ['gr_LTA1'], # lipoteichoic acids
        'LTAga40g_20t_20a_c':  ['gr_LTA2'], # lipoteichoic acids 
        'LTAt4_20a_c': ['gr_LTA3'],  # lipoteichoic acids 
        'peptidoSTA_c': ['gr_ptdSTA'], # peptidoglycan (dependant on 'udcpdp_c')
        'peptidoSTR_c': ['gr_ptdSTR'],
        'peptidoDAP_c': ['gr_ptdDAP'],
        'epsLAB_c': ['gr_epsLAB'],  # EPS of lactic acid bacteria
        'phb_c': ['gr_PHA1'],  # PHA / PHB    
    }
    
    
    return precursors_to_pathway



def introduce_universal_biomass(logger, dbexp, universe): 
    
    
    biomass_dict = get_biomass_dict(logger, universe, dbexp)
    
    
    # MWF: macromolecular weight fractions:
    rstring =              f'{get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["RNA"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["DNA"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["PROTS"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["LIPIDS_PL"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["LIPIDS_FA"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["WTA"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["LTA"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["PEPTIDO"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["EPS"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["PHA"])}'
    
    
    # Cofactors:
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["cofs_uni_Xavier2017"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["cofs_cond_Xavier2017"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(biomass_dict["cofs_suppl"])}'
    rstring = rstring + f' + {get_uni_biom_coeff()} atp_c + {get_uni_biom_coeff()} h2o_c --> {get_uni_biom_coeff()} adp_c + {get_uni_biom_coeff()} h_c + {get_uni_biom_coeff()} pi_c'
    
    
    r = cobra.Reaction('Biomass')
    universe.add_reactions([r])
    r = universe.reactions.get_by_id('Biomass')
    r.name = 'Biomass assembly reaction'
    r.build_reaction_from_string(rstring)
    
    
    # add SBO annotation
    r.annotation['sbo'] = ['SBO:0000629']  # biomass reaction 
    
    
    # set as objective:
    universe.objective = 'Biomass'
    
    
    return universe



def adjust_biomass_precursors(logger, model, universe, conditional_threshold):
    

    precursor_to_pathway = get_precursors_to_pathway()
    modeled_rids = [r.id for r in model.reactions]
    
    
    cnt_removed = 0
    cond_col_dict = {}  # dictionary of variable biomass precursors
    for precursor, pathways in precursor_to_pathway.items(): 
        
        pathway_to_absence = {}
        pathway_to_compstring = {}   # completeness string
        for pathway in pathways:   # more pathways might lead to the same precursor
            # initialize counters:
            cnt_members_tot = 0
            cnt_members_present = 0

            
            if pathway not in [gr.id for gr in universe.groups]:
                continue   # still missing from the universe
                
            for member in universe.groups.get_by_id(pathway).members:
                cnt_members_tot += 1    
                if member.id in modeled_rids:
                    cnt_members_present += 1
            # populate dicts:
            pathway_to_absence[pathway] = (cnt_members_present / cnt_members_tot) < conditional_threshold
            pathway_to_compstring[pathway] = f'{pathway}: {cnt_members_present}/{cnt_members_tot}'
            

        cond_col_dict[precursor] = '; '.join(list(pathway_to_compstring.values()))
        if all(list(pathway_to_absence.values())):
            cnt_removed += 1
            logger.debug(f"Biomass precursor '{precursor}' seems not required ({cond_col_dict[precursor]}).")
            # add metabolites to the right side (they will disappear if the balance if 0)
            model.reactions.Biomass.add_metabolites({precursor: get_uni_biom_coeff()})

       
    if model.id != universe.id:   # avoid logging when testing blocked biomass precusors of the universe.
        logger.info(f"Removed {cnt_removed} conditional biomass precursors.")
        
        
    return cond_col_dict



def compute_biomass_norm(logger, model, biomass): 
    
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # avoids warnings like "UserWarning: 9.823 is not an integer (in formula C9.823H5.5789999999999935N3.6769999999999996O3.628P1.0)"
    
    
        norm = Decimal(0.0)
        for m, coeff in model.reactions.get_by_id('Biomass').metabolites.items():
            if coeff < 0.0:   # reactant
                if m.id not in ["atp_c", "h2o_c"]:
                    # '/1000' as they are milli-model coefficients:
                    norm = norm + Decimal(abs(coeff)) / 1000 * Decimal(m.formula_weight) 
        logger.info(f"Biomass adjusted using '{biomass}' data: {str(round(norm, 16))} gDW.")



def adjust_biomass_coefficients(logger, model, universe, dbexp, biomass):
    # Strategy: the default coefficients are equal to 'get_uni_biom_coeff()', which is  a really samm number, negligible.
    # Coefficients are adjusted only for precursors comprised by MWF, all the others are left with default coefficients.
    # Therefore, if the final biomass is not sufficiently close to 1.0, just lower the 'get_uni_biom_coeff()', 
    # or lower the number of precursors left with default coefficient. 
    
    
    if biomass == '-':
        logger.info(f"No biomass data provided: using minimal coefficients ({get_uni_biom_coeff()}).")
        return 0
    
    
    # check biomass data
    response = check_exp_biomass_data(logger, dbexp, biomass)
    if type(response)==int: return 1
    else: MWF, DNA, RNA, PROTS, LIPIDS_PL, LIPIDS_FA = response
    
    
    # compute coefficient for 'Biomass' reaction:
    biomass_dict = get_biomass_dict(logger, universe, dbexp)
    reacs_prods_coeffs = dict()
    
    
    # the reaction 'MEANFA' will replace 'sn_meanfa_c' to produce 'meanfa_c'.
    model.remove_reactions([model.reactions.get_by_id("sn_meanfa_c")])
    logger.debug(f"Removed 'sn_meanfa_c': will be replaced by 'MEANFA'.")
    
    
    # DNA
    response = compute_exp_DNA_coeffs(logger, model, MWF, DNA)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['DNA'] = response
    # RNA
    response = compute_exp_RNA_coeffs(logger, model, MWF, RNA)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['RNA'] = response
    # PROTS
    response = compute_exp_PROTS_coeffs(logger, model, MWF, PROTS)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['PROTS'] = response
    # LIPIDS
    response = compute_exp_LIPIDS_coeffs(logger, model, MWF, LIPIDS_PL, LIPIDS_FA)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['LIPIDS'] = response 
    # WTA
    response = compute_exp_WTA_coeffs(logger, model, biomass_dict, MWF, dbexp)    
    if type(response)==int: return 1
    else: reacs_prods_coeffs['WTA'] = response
    # LTA
    response = compute_exp_LTA_coeffs(logger, model, biomass_dict, MWF, dbexp)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['LTA'] = response
    # PEPTIDO
    response = compute_exp_PEPTIDO_coeffs(logger, model, biomass_dict, MWF, dbexp)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['PEPTIDO'] = response
    # EPS
    response = compute_exp_EPS_coeffs(logger, model, biomass_dict, MWF, dbexp)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['EPS'] = response  
    
    
    # merge dicts of reactants and products: 
    reactant_to_coeffs = dict()
    product_to_coeffs = dict()
    for i in reacs_prods_coeffs.values():
        reacs, prods = i[0], i[1]
        for key, coeff in reacs.items():
            if key not in reactant_to_coeffs.keys():
                reactant_to_coeffs[key] = Decimal('0.0')
            reactant_to_coeffs[key] = reactant_to_coeffs[key] + coeff
        for key, coeff in prods.items():
            if key in ['DNA_c', 'RNA_c', 'PROTS_c', 'LIPIDS_c']:
                continue
            if key not in product_to_coeffs.keys():
                product_to_coeffs[key] = Decimal('0.0')
            product_to_coeffs[key] = product_to_coeffs[key] + coeff
    
    
    # update biomass coefficients
    for mid in [m.id for m in model.reactions.get_by_id('Biomass').reactants]:
        if mid in reactant_to_coeffs.keys(): 
            model.reactions.Biomass.add_metabolites({mid: get_uni_biom_coeff()})
            model.reactions.Biomass.add_metabolites({mid: -float(reactant_to_coeffs[mid])})
    for mid in [m.id for m in model.reactions.get_by_id('Biomass').products]:
        if mid in product_to_coeffs.keys(): 
            model.reactions.Biomass.add_metabolites({mid: -get_uni_biom_coeff()})
            model.reactions.Biomass.add_metabolites({mid: float(product_to_coeffs[mid])})
            
            
    # remove unsed biomass precursors:
    biomass_dict = get_biomass_dict(logger, universe, dbexp)
    # remove all fatty acids:
    for mid in biomass_dict["LIPIDS_FA"]:
        model.reactions.Biomass.add_metabolites({mid: get_uni_biom_coeff()})
    # remove unused phospholipids:
    for mid in biomass_dict["LIPIDS_PL"]:
        if mid not in reacs_prods_coeffs['LIPIDS'][0].keys():
            model.reactions.Biomass.add_metabolites({mid: get_uni_biom_coeff()})
            
            
    # check normalization to 1.0 gDW
    compute_biomass_norm(logger, model, biomass)
            
    
    return 0



def get_biomass_df_structure(logger, df, model, universe, dbexp, cond_col_dict):
    
    
    # get strain-specific biomass precursors / coeffs to gap-fill: 
    biomass_mids_to_coeff = dict()
    for m, coeff in model.reactions.Biomass.metabolites.items():
        if coeff < 0:   # reactants
            biomass_mids_to_coeff[m.id] = abs(coeff)
            
            
    # get all universal biomass precursors: 
    biomass_dict = get_biomass_dict(logger, universe, dbexp)
    biomass_mids = []
    for key, values in biomass_dict.items():
        for i in values:
            biomass_mids.append(i)
            
    
    # preload ctegory of each precursor
    precursor_to_catg = dict()
    for catg, mids in biomass_dict.items(): 
        for mid in mids: 
            precursor_to_catg[mid] = catg
    precursor_to_catg['atp_c'] = '/'
    
    
    
    # prepare sheet for excel output
    df['name'] = None
    df['category'] = None
    df['cond'] = None
    df['coeff'] = None
    universal_precs = ['atp_c'] + [m.id for m in universe.reactions.Biomass.reactants if m.id not in ['atp_c', 'h2o_c']]
    for mid in universal_precs:
        
        if mid in biomass_mids and not mid in biomass_mids_to_coeff.keys():
            if mid in biomass_dict['LIPIDS_PL'] or mid in biomass_dict['LIPIDS_FA']:
                continue   # unobserved experimentally, not true conditional 
        
        df.loc[mid, 'name'] = universe.metabolites.get_by_id(mid).name
        
        df.loc[mid, 'category'] = precursor_to_catg[mid]
        
        # set 'cond' column:
        if mid in cond_col_dict.keys():
            df.loc[mid, 'cond'] = f'{cond_col_dict[mid]}'
        else:
            df.loc[mid, 'cond'] = '/'
            
        # set 'coeff' column:
        if mid not in [m.id for m in model.reactions.Biomass.reactants]:  
                df.loc[mid, 'coeff'] = 'removed'
        else:
            df.loc[mid, 'coeff'] = biomass_mids_to_coeff[mid]
            
            
    return df








def force_id_on_sbml(file_path, model_id):
    
    with open(file_path, 'r') as file:
        content = file.read()
    content = content.replace(
        f'<model metaid="meta_{model_id}" fbc:strict="true">', 
        f'<model metaid="meta_{model_id}" id="{model_id}" fbc:strict="true">'
    )
    with open(file_path, 'w') as file:
        file.write(content)


import warnings 
import os
import logging


import cobra
import gempipe


from ..commons import fba_no_warnings
from ..commons import get_optthr



def verify_egc(logger, model, mid, outdir): 

    
    # changes as not permament: 
    found_egc = False
    with model: 
        
        # close (0; 0) all the exchange reactions: 
        gempipe.close_boundaries(model)
        
                
        # create a dissipation reaction: 
        dissip = cobra.Reaction(f'__dissip__{mid}')
        model.add_reactions([dissip])
        dissip = model.reactions.get_by_id(f'__dissip__{mid}')
        
        
        # define the dissipation reaction:
        modeled_mids = [m.id for m in model.metabolites]
        if   mid == 'atp':
            dissip_string = 'atp_c + h2o_c --> adp_c + pi_c + h_c'
        elif mid == 'ctp':
            dissip_string = 'ctp_c + h2o_c --> cdp_c + pi_c + h_c'
        elif mid == 'gtp':
            dissip_string = 'gtp_c + h2o_c --> gdp_c + pi_c + h_c'
        elif mid == 'utp':
            dissip_string = 'utp_c + h2o_c --> udp_c + pi_c + h_c'
        elif mid == 'itp':
            dissip_string = 'itp_c + h2o_c --> idp_c + pi_c + h_c'
        elif mid == 'nadh':
            dissip_string = 'nadh_c --> nad_c + h_c'
        elif mid == 'nadph':
            dissip_string = 'nadph_c --> nadp_c + h_c'
        elif mid == 'fadh2':
            dissip_string = 'fadh2_c --> fad_c + 2.0 h_c'
        elif mid == 'accoa':
            dissip_string = 'accoa_c + h2o_c --> ac_c + coa_c + h_c'
        elif mid == 'glu__L':
            dissip_string = 'glu__L_c + h2o_c --> akg_c + nh4_c + 2.0 h_c'
        elif mid == 'q8h2':
            dissip_string = 'q8h2_c --> q8_c + 2.0 h_c'
        dissip.build_reaction_from_string(dissip_string)
        
        
        # set the objective and optimize: 
        model.objective = f'__dissip__{mid}'
        res, obj_value, status = fba_no_warnings(model)
        
        
        # apply the threshold:
        obj_value = res.objective_value
        status = res.status
        if status == 'optimal' and obj_value >= get_optthr():
            found_egc = True
                

            # get suspect !=0 fluxes 
            fluxes = res.fluxes
            # get interesting fluxes (get_optthr() tries to take into account the approximation in glpk and cplex solvers)
            fluxes_interesting = fluxes[(fluxes > get_optthr()) | (fluxes < -get_optthr())]
            

            # create a model for escher, remove Rs not beloning to the cycle
            model_copy = model.copy()
            all_rids = [r.id for r in model_copy.reactions]
            to_delete = set(all_rids) - set(fluxes_interesting.index)


            # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
            # triggered when trying to remove reactions that are included in groups. 
            with warnings.catch_warnings():  # temporarily suppress warnings for this block
                warnings.simplefilter("ignore")  # ignore all warnings
                cobra_logger = logging.getLogger("cobra.util.solver")
                old_level = cobra_logger.level
                cobra_logger.setLevel(logging.ERROR)   
                
                # triggering code
                model_copy.remove_reactions(to_delete)  # should work also with IDs

                # restore original behaviour: 
                cobra_logger.setLevel(old_level) 
                
                
            # save JSON to direct import in Escher:
            outfile = os.path.join(outdir, f'EGC_{mid}.json')
            cobra.io.save_json_model(model_copy, outfile)


            # log some messages
            rid_labels = []
            for rid, flux in fluxes_interesting.to_dict().items():
                rid_label = "'" + rid + "'" 
                # mark reversible reactions composing the cycle:
                r = model.reactions.get_by_id(rid)
                if r.lower_bound < 0 and r.upper_bound > 0:
                    rid_label = rid_label + '(<=>)'
                rid_labels.append(rid_label)
            logger.warning(f"Found erroneous EGC (N={len(model_copy.reactions)}) for '{mid}' (f={obj_value}): [{', '.join(rid_labels)}]. EGC saved to '{outfile}' to be inspected with Escher-FBA.")
            
            
    return found_egc



def verify_egc_all(logger, model, outdir='./', mids_to_check=['atp','ctp','gtp','utp','itp','nadh','nadph','fadh2','accoa','glu__L','q8h2']):

    
    all_results = []
    for mid in mids_to_check:
        all_results.append(verify_egc(logger, model, mid, outdir))
    if any(all_results)==False:
        logger.info("Found 0 erroneous energy-generating cycles (EGCs).")
        
        



from io import BytesIO

import numpy as np
import pandas as pnd

from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage, cut_tree, dendrogram, leaves_list

import matplotlib.pyplot as plt
from matplotlib.patches import Patch



def figure_df_C_F1(df_coverage):

    
    
    # prepare the binary matrix: 
    modeled_rs = df_coverage[df_coverage['modeled']==True].index
    unmodeled_rs = df_coverage[df_coverage['modeled']==False].index
    # remove useless columns
    bin_matrix = df_coverage[[i for i in df_coverage.columns if i not in ['map_ids', 'modeled']]]
    # sort rows: upper rows are present in more strains
    bin_matrix = bin_matrix.loc[bin_matrix.sum(axis=1).sort_values(ascending=False).index]
    # split in 2: modeled above, non-modeled below:
    bin_matrix = pnd.concat([
        bin_matrix.loc[[i for i in bin_matrix.index if i in modeled_rs], ], 
        bin_matrix.loc[[i for i in bin_matrix.index if i in unmodeled_rs], ]
    ])  
    strains = bin_matrix.columns
    bin_matrix = bin_matrix.T  # features in column
    

    # pdist() / linkage() will loose the accession information. So here we save a dict: 
    index_to_strain = {i: strain for i, strain in enumerate(bin_matrix.index)}

    # Calculate the linkage matrix using Ward clustering and Jaccard dissimilarity
    distances = pdist(bin_matrix, 'jaccard')
    linkage_matrix = linkage(distances, method='ward')


    # PART 0: create the frame
    fig, axs = plt.subplots(
        nrows=2, ncols=2, 
        figsize=(15, 10), 
        gridspec_kw={  # suplots width proportions. 
            'width_ratios': [0.5, 1.0],
            'height_ratios': [0.015, 0.985]
        }
    ) 

    # PART 1: dendrogram
    dn = dendrogram(
        linkage_matrix, ax=axs[1,0],
        orientation='left',
        color_threshold=0, above_threshold_color='black',
    )


    ### PART 2: heatmap
    ord_leaves = leaves_list(linkage_matrix)
    ord_leaves = np.flip(ord_leaves)  # because leaves are returned in the inverse sense.
    ord_leaves = [index_to_strain[i] for i in ord_leaves]  # convert index as number to index as accession
    bin_matrix = bin_matrix.loc[ord_leaves, :]  # reordered dataframe.
    axs[1,1].matshow(
        bin_matrix,  
        cmap='viridis',
        aspect='auto', # non-squared pixels to fit the axis
    )


    ### PART 3: coverage bar
    axs[0,1].matshow(
        df_coverage.loc[bin_matrix.T.index, ['modeled']].T,  
        cmap='cool_r',
        aspect='auto', # non-squared pixels to fit the axis
    )


    ### PART 4: legends
    legend_feat = [
        Patch(facecolor=plt.colormaps.get_cmap('viridis')(0.0), edgecolor='black', label='Absent'),
        Patch(facecolor=plt.colormaps.get_cmap('viridis')(1.0), edgecolor='black', label='Probably present'),
    ]
    legend_cov = [
        Patch(facecolor=plt.colormaps.get_cmap('cool_r')(0.0), edgecolor='black', label='Not modeled'),
        Patch(facecolor=plt.colormaps.get_cmap('cool_r')(1.0), edgecolor='black', label='Modeled'),
    ]
    l1 = axs[1,0].legend(handles=legend_cov, title='Universe coverage', loc='upper left')
    l2 = axs[1,0].legend(handles=legend_feat, title='KEGG reaction in strain', loc='lower left')
    axs[1,0].add_artist(l1)  # keep both legends visible


    ### PART 5: aesthetics
    plt.subplots_adjust(wspace=0, hspace=0)  # adjust the space between subplots: 
    axs[0,0].axis('off')  # remove frame and axis
    axs[1,0].axis('off')  # remove frame and axis

    axs[0,1].yaxis.set_visible(False)  # remove ticks, tick labels, axis label

    axs[1,1].xaxis.set_ticks([])       # remove ticks
    axs[1,1].set_xticklabels([])       # remove tick labels
    axs[1,1].xaxis.set_label_position("bottom")
    axs[1,1].set_xlabel("KEGG reactions")

    axs[1,1].yaxis.set_ticks([])       # remove ticks
    axs[1,1].set_yticklabels([])       # remove tick labels
    axs[1,1].yaxis.set_label_position("right")
    axs[1,1].set_ylabel(f"{len(strains)} strains", rotation=270, labelpad=13)  # labelpad is in points (1 point = 1/72 inch)


    ### PART 6: save fig
    buf = BytesIO()
    fig.savefig(buf, dpi=300, bbox_inches='tight')  # labelpad is in inches (1 point = 1/72 inch)
    plt.close(fig)
    buf.seek(0)  # rewind the buffer to the beginning
    
    
    return buf



import os
import contextlib
import importlib.metadata 



import memote




def get_memote_results_dict(logger, model):
    
    
    logger.info(f"Running selected modules of MEMOTE v{importlib.metadata.metadata('memote')['Version']}...")


    # launch memote (only relevant modules)        
    with open(os.devnull, 'w') as devnull:
        with contextlib.redirect_stdout(devnull), contextlib.redirect_stderr(devnull):
            try: memote_report = memote.suite.api.test_model(model, exclusive=[
                'test_annotation', 
                'test_sbo',
                'test_stoichiometric_consistency',
                'test_reaction_mass_balance',
                'test_reaction_charge_balance',
                'test_find_disconnected',
                'test_find_reactions_unbounded_flux_default_condition'], results=True)
            except ValueError: memote_report = None
            

    # parse memote's results
    results_dict = {}
    results_dict['version'] = importlib.metadata.version("memote")
    test_results = dict(memote_report[1])['tests']
    sections = {
        'consistency': [
            ('test_stoichiometric_consistency', 3),
            ('test_reaction_mass_balance', 1),
            ('test_reaction_charge_balance', 1),
            ('test_find_disconnected', 1),
            ('test_find_reactions_unbounded_flux_default_condition', 1)
        ],
        'annotation_M': [
            ('test_metabolite_annotation_presence', 1),
            ('test_metabolite_annotation_overview', 1),
            ('test_metabolite_annotation_wrong_ids', 1),
            ('test_metabolite_id_namespace_consistency', 1),
        ],
        'annotation_R': [
            ('test_reaction_annotation_presence', 1),
            ('test_reaction_annotation_overview', 1),
            ('test_reaction_annotation_wrong_ids', 1),
            ('test_reaction_id_namespace_consistency', 1),
        ],
        'annotation_G': [
            ('test_gene_product_annotation_presence', 1),
            ('test_gene_product_annotation_overview', 1),
            ('test_gene_product_annotation_wrong_ids', 1),
        ],
        'annotation_SBO': [
            ('test_metabolite_sbo_presence', 1),
            ('test_metabolite_specific_sbo_presence', 1),
            ('test_reaction_sbo_presence', 1),
            ('test_metabolic_reaction_specific_sbo_presence', 1),
            ('test_transport_reaction_specific_sbo_presence', 1),
            ('test_exchange_specific_sbo_presence', 1),
            ('test_demand_specific_sbo_presence', 1),
            ('test_sink_specific_sbo_presence', 1),
            ('test_gene_sbo_presence', 1),
            ('test_gene_specific_sbo_presence', 1),
            ('test_biomass_specific_sbo_presence', 1),
        ],
    }
    section_multipliers = {
        'consistency': 3,
        'annotation_M': 1,
        'annotation_R': 1,
        'annotation_G': 1,
        'annotation_SBO': 2,
    }
    
    
    numerator_total = 0
    denominator_total = 0
    for section, metrics in sections.items(): 
        numerator = 0
        denominator = 0
        results_dict[section] = {}
        
        
        # iterate metrics of this section: 
        for metric, metric_multiplier in metrics:
            metric_raw = test_results[metric]['metric']
            
            
            # no subcategories here: 
            if type(metric_raw) == float: 
                metric_percentage = ((1- metric_raw ) *100)
                numerator = numerator + (metric_percentage * metric_multiplier)
                denominator = denominator + metric_multiplier
                results_dict[section][metric] = round(metric_percentage, 1)
                                
            
            # there are subcategories (like in the case of M/R/G/SBO annots)
            else:   
                results_dict[section][metric] = {}
                for key, value in metric_raw.items():
                    n_subcategories = len(metric_raw)
                    multiplier_corrected = metric_multiplier / n_subcategories
                    metric_percentage = ((1- value ) *100)
                    numerator = numerator + (metric_percentage * multiplier_corrected)
                    denominator = denominator + multiplier_corrected
                    results_dict[section][metric][key] = round(metric_percentage, 1)
                    
                    
        # compute the subtotal:
        sub_total = numerator / denominator
        results_dict[section]['sub_total'] = int(round(sub_total, 0))
        
        
        # compute the total:
        denominator_total = denominator_total + section_multipliers[section] *denominator
        numerator_total = numerator_total + section_multipliers[section] *numerator 
    total = numerator_total / denominator_total
    results_dict['total'] = int(round(total, 0))


    logger.info(f"Done! MEMOTE Total Score: {results_dict['total']}%.")
    
    
    return results_dict





def get_deprecated_kos():
    deprecated_kos = [
        'K11189',  # should be K02784
        'K07011',  # linked to lp_1215(cps3A) and lp_1216(cps3B) during 2018 and not replaced
    ]
    return deprecated_kos



def get_rids_with_mancheck_gpr():
    rids_mancheck_gpr = [  # reactions with manually checked GPRs
        'SUCD1', 'ALKP', 'PFK_3', 'TCMPTS', 'PPA', 'APSR',
        'NADHPO', 'ACOD', '5DOAN', '5DRBK', '5DRBPI', '5DRBUPD',
        'FNPOR', 'THZPSN', 'PPTGP', 'ALATA_D2', 'XYLUR1'
    ]
    return rids_mancheck_gpr


def get_rids_with_mancheck_balancing():
    rids_mancheck_bal = [  # same reactions involving ATP can be reversible
        
        # SECTION "reversible both in KEGG and MetaCyc"
        'PGK', 'SUCOAS', 'ADK1', 'GK1', 'NNATr', 'CYTK1', 'ACKr',
        'DGK1', 'PPAKr', 'ATPSr', 'NDPK10',
        
        ### SECTION "reversible in KEGG but not in MetaCyc" ###
        'CYTK2',  # clearly reversible in KEGG but not in MetaCyc (RXN-7913)
        'DADK',  # clearly reversible in KEGG but not in MetaCyc (DEOXYADENYLATE-KINASE-RXN)
        'UMPK',  # clearly reversible in KEGG but not in MetaCyc (RXN-12002)
        'NDPK1',  # clearly reversible in KEGG but not in MetaCyc (GDPKIN-RXN)
        'NDPK2',  # clearly reversible in KEGG but not in MetaCyc (UDPKIN-RXN)  
        'NDPK3',  # clearly reversible in KEGG but not in MetaCyc (CDPKIN-RXN)
        'NDPK4',  # clearly reversible in KEGG but not in MetaCyc (DTDPKIN-RXN)
        'NDPK5',  # clearly reversible in KEGG but not in MetaCyc (DGDPKIN-RXN)
        'NDPK6',  # clearly reversible in KEGG but not in MetaCyc (DUDPKIN-RXN)
        'NDPK7',  # clearly reversible in KEGG but not in MetaCyc (DCDPKIN-RXN)
        'NDPK8',  # clearly reversible in KEGG but not in MetaCyc (DADPKIN-RXN)
        'NDPK9',  # clearly reversible in KEGG but not in MetaCyc (RXN-14120) 
        
        ### SECTION "missing reversibility info" ###
        'LPHERA',  
    ]
    return rids_mancheck_bal



def get_manual_sinks():
    
    return ['apoACP', 'aponit', 'apocarb', 'thioca', 'THI5p_b', 'cyE', 'meanfa']



def get_manual_demands():
    
    return ['scp', 'amob', 'dialurate', 'THI5p_a', 'partmass' ]



def get_custom_groups():
    
    
    return {
        'gr_ptdSTA': ['UAMAGLL', 'UMNS', 'UPPNAPT', 'UAGPT2', 'UAGPGAL', 'UAGPN6GT', 'UAAGGTLGAGT', 'UAAGGTLG3AGT', 'PPTGP'],
        'gr_ptdSTR': ['UAMAGLL', 'UMNS', 'UPPNAPT', 'UAGPT2', 'PPGAAE1', 'PPGAAE2', 'PPTGP2'],
        'gr_ptdDAP': ['UGMDL', 'UGMDDS', 'UGMDDPT', 'UAGPT3', 'PPTGP3'],
        'gr_HemeO': ['HEMEOS'],
        'gr_WTA1': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTAGPP', 'WTAUGLCT2', 'WTAALAT3', 'WTAPL3'],
        'gr_WTA2': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTARBT2', 'WTARPP2', 'WTAUGLCT', 'WTAALAT2', 'WTAPL2'],
        'gr_WTA3': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTAGPT2', 'WTARPP', 'WTAGLCNACT', 'WTAALAT', 'WTAPL'],
        'gr_WTA4': ['UNACGYL', 'UADDTRS', 'AATGALT', 'TAA13GLT', 'TAARPTR', 'TAANGTR', 'TAANGTR2', 'LPSCHPT', 'TACCHPT', 'T4WTAPOL', 'WTAPL4', 'WTAT4ALAT'], # type-IV WTA
        'gr_LTA1': ['UGDIAT', 'UGLDIAT', 'GGGDAGF2', 'LIPOPO2', 'LTANACT', 'LTAALAT2'],
        'gr_LTA2': ['UGDIAT', 'UGADIAT', 'GGGDAGF', 'LIPOPO', 'LTAGAT', 'LTAALAT'],
        'gr_LTA3': ['UNACGYL', 'UADDTRS', 'AATGALT', 'TAA13GLT', 'TAARPTR', 'TAANGTR', 'TAANGTR2', 'LPSCHPT', 'TACCHPT', 'T4WTAPOL', 'T4LTAL', 'LTAT4ALAT'], # type-IV LTA
        'gr_br': ['LYEH1', 'HIPCD1', 'LYEH2', 'HIPCD2', 'BHBRH1', 'BHBRH2'],
        'gr_PHA1': ['ACACT1r', 'AACOAR_syn', 'PHBS_syn_1', 'PHBDEP_1'],    # PHA from glycolyis
        'gr_epsLAB': ['PGTLAB', 'GTSTLAB', 'EPSPOLAB'],
    }


import warnings
import logging



def remove_forced(logger, model, universe, force_removal):
    
    
    if force_removal != '-':
        removed_rids = []
        forced_rids = force_removal.split(',')
        
        modeled_rids = [r.id for r in model.reactions]
        universal_rids = [r.id for r in universe.reactions]
        
        for rid in forced_rids: 
            
            if rid not in universal_rids:
                logger.info(f"Ignoring reaction ID '{rid}' since it's not included in the universe.")
                continue
            
            if rid not in modeled_rids:
                logger.debug(f"Requested reaction ID '{rid}' was already excluded from the model.")
            else:
                removed_rids.append(rid)

        # remove collected reactions:
        with warnings.catch_warnings():
            # avoid warnings like "python3.9/site-packages/cobra/core/group.py:147: UserWarning: need to pass in a list"
            model.remove_reactions(removed_rids) # this works also with rids instaead of Rs

        logger.info(f"Reactions forcibly removed: {removed_rids}.")
        


def remove_disconnected(logger, model):
    
    
    to_remove = []
    for m in model.metabolites:
        if len(m.reactions) == 0:
            to_remove.append(m)
    model.remove_metabolites(to_remove)
    logger.info(f"Removed {len(to_remove)} disconnected metabolites.")
    
    
    return model

    
    
def remove_universal_orphans(logger, model):
    
    
    to_remove = []
    for r in model.reactions:
        if len(r.metabolites) != 1 and r.id != 'Biomass':  # exclude artificials
            if r.gene_reaction_rule == '': 
                logger.debug(f"Removing universal orphan '{r.id}'...")
                to_remove.append(r.id)
            
    
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        model.remove_reactions(to_remove) 
        logger.info(f"Removed {len(to_remove)} universal orphan reactions.")

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
            

    return model
    
    
    
def remove_sinks_demands(logger, model): 
    
    
    to_remove = []
        
        
    for r in model.reactions: 
        if len(r.metabolites) == 1:    # sinks / damands / exchanges
            m = list(r.metabolites)[0]
            if m.id.endswith('_e') == False:   # sinks / damands
                if len(m.reactions) == 1 and list(m.reactions)[0].id == r.id:
                    # if it's associated only with this reaction:
                    logger.debug(f"Removing unused sink/demand '{r.id}'...")
                    to_remove.append(r.id)
                    
    model.remove_reactions(to_remove) 
    logger.info(f"Removed {len(to_remove)} disconnected sink/demand reactions.")
    
    
    return model


import os
from pathlib import Path
import logging
import tempfile
import concurrent.futures as confu 
import sys


import cobra
import gempipe


from .pruner import load_input_universe
from .pruner import load_input_eggnog
from .pruner import load_keggorg_like_eggnog
from .pruner import parse_eggnog
from .pruner import subtract_kos
from .pruner import translate_remaining_kos
from .pruner import restore_gene_annotations
from .pruner import append_keggorg_gene_annots

from .gapfillutils import include_forced

from .gapfill import gapfill_on_media

from .polishing import remove_universal_orphans
from .polishing import remove_forced
from .polishing import remove_disconnected
from .polishing import remove_sinks_demands

from .biologcuration import biolog_on_media

from ..commons import get_logger
from ..commons import get_databases
from ..commons import adjust_biomass_precursors
from ..commons import adjust_biomass_coefficients
from ..commons import force_id_on_sbml
from ..commons import write_excel_model
from ..commons import log_metrics
from ..commons import log_unbalances
from ..commons import format_expansion
from ..commons import comparative_table
from ..commons import download_keggorg
from ..commons import initialize_model
from ..commons import get_memote_results_dict


from ..runsims.biosynth import biosynthesis_on_media
from ..runsims.simplegrowth import grow_on_media

from ..parsedb.cycles import verify_egc_all




def create_model_incore(params):
    annotation_source, universe, eggpath, dbexp, args, multistrain = params
    
    # get the logger:
    logger = get_logger('gsrap_queued', args.verbose)  # loggers can't be pickled!
    
    
    # only errors will be recorded if multistrain mode
    if multistrain:
        logger.setLevel(logging.ERROR)  


    # load the annotation
    if annotation_source == 'keggorg':
        eggnog_style_table = load_keggorg_like_eggnog(logger, args.keggorg, args.outdir)
    elif annotation_source == 'eggnog':
        eggnog_style_table = load_input_eggnog(logger, eggpath)


    # create a copy of the universe and define the model ID
    model = universe.copy()
    if annotation_source == 'keggorg':
        model.id = args.keggorg
    elif annotation_source == 'eggnog':
        model.id = Path(eggpath).stem 


    ###### POLISHING 1
    # remove universal orphans
    model = remove_universal_orphans(logger, model)

    

    ###### PRUNING
    if   annotation_source == 'keggorg': logger.info(f"Reading annotation for organism code '{args.keggorg}'...")
    elif annotation_source == 'eggnog':  logger.info("Reading provided eggnog-mapper annotation...")
    # get important dictionaries: 'eggnog_ko_to_gids' and 'eggonog_gid_to_kos'
    eggnog_ko_to_gids, eggonog_gid_to_kos = parse_eggnog(eggnog_style_table)    

    # prune reactions
    subtract_kos(logger, model, eggnog_ko_to_gids)

    # translate KOs to the actual genes
    translate_remaining_kos(logger, model, eggnog_ko_to_gids)
    restore_gene_annotations(logger, model, universe, eggonog_gid_to_kos)
    
    # insert gene annotation if starting from kegg organisms:
    if annotation_source == 'keggorg': 
        append_keggorg_gene_annots(logger, model, args.keggorg, args.outdir)

    

    ###### GAPFILLING
    # force inclusion of reactions:  
    include_forced(logger, model, universe, args.include)

    # remove missing conditional precursors + get the 'cond_col_dict' dict.
    # 'cond_col_dict' is str-to-str: {'pheme_c': 'M00868: 1/8; M00121: 2/12;', 'hemeO_c': 'gr_HemeO: 0/1'}
    cond_col_dict = adjust_biomass_precursors(logger, model, universe, args.conditional) 

    # adjust biomass coefficients
    response = adjust_biomass_coefficients(logger, model, universe, dbexp, args.biomass)
    if response == 1: return 1

    # gap-fill based on media:
    df_B = gapfill_on_media(logger, model, universe, dbexp, args.gapfill, cond_col_dict, args.excludeorp)
    if type(df_B)==int: return 1

    # force removal of reactions
    setattr(args, 'remove', '-')  # experimental feature, not public. It's main purpose was to test gap-filling in biolog_on_media().
    remove_forced(logger, model, universe, args.remove)

    # perform Biolog(R) curation based on media
    df_P = biolog_on_media(logger, model, universe, dbexp, args.gapfill, args.biolog, args.excludeorp, args.cnps)
    if type(df_P)==int: return 1



    ###### POLISHING 2
    # remove unsed sinks and demands
    model = remove_sinks_demands(logger, model)  
    
    # remove disconnected metabolites
    model = remove_disconnected(logger, model)

    

    # # # # #   DERIVATION ENDS HERE   # # # # #
    log_metrics(logger, model)
    log_unbalances(logger, model)


    
    ###### CHECKS 1
    # check erroneous EGCs
    verify_egc_all(logger, model, args.outdir)
    
    # check blocked metabolites / dead-ends
    df_S = biosynthesis_on_media(logger, model, dbexp, args.gapfill, args.biosynth)
    if type(df_S)==int: return 1



    ###### POLISHING 3
    # reset growth environment befor saving the model
    gempipe.reset_growth_env(model)
    
    # initialize model
    response = initialize_model(logger, model, dbexp, args.initialize, args.gapfill)
    if response==1: return 1



    ###### CHECKS 2
    # compute Memote metrics
    memote_results_dict = get_memote_results_dict(logger, model)


    
    # output the model:
    logger.info("Writing strain-specific model...")
    cobra.io.save_json_model(model, f'{args.outdir}/{model.id}.json')        # JSON
    logger.info(f"'{args.outdir}/{model.id}.json' created!")
    cobra.io.write_sbml_model(model, f'{args.outdir}/{model.id}.xml')        # SBML   # groups are saved only to SBML
    logger.info(f"'{args.outdir}/{model.id}.xml' created!")
    force_id_on_sbml(f'{args.outdir}/{model.id}.xml', model.id)   # force introduction of the 'id=""' field
    sheets_dict = write_excel_model(model, f'{args.outdir}/{model.id}.mkmodel.xlsx', args.nofigs, memote_results_dict, None, df_B, df_P, df_S)  
    logger.info(f"'{args.outdir}/{model.id}.mkmodel.xlsx' created!")
        
        
    return sheets_dict    



def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # determine available cores
    cpu_count = os.cpu_count()
    if cpu_count == None: 
        logger.debug("Failed to count cpus. Reset to 1.")
        cpu_count = 1  
    if args.cores == 0:
        args.cores = cpu_count
    elif args.cores > cpu_count:
        logger.debug(f"Parameter --cores {args.cores} is greater than the maximum available ({cpu_count}): reset to {cpu_count}.")
        args.cores = cpu_count
    
    
    # format the --eggnog param
    args.eggnog = format_expansion(logger, args.eggnog)  # now 'args.eggnog' could still be '-'
    
    # get the kegg organism if requested
    if args.keggorg != '-':
        response = download_keggorg(logger, args.keggorg, args.outdir)
        if response == 1: return 1
        
    
    
    # determine the source of functional annotation:
    annotation_source = None
    if args.keggorg != '-':  # keggorg has precedence
        annotation_source = 'keggorg'
    elif args.eggnog != '-':
        annotation_source = 'eggnog'
        if args.cores > len(args.eggnog):
            logger.debug(f"Parameter --cores {args.cores} is greater than the number of strains ({len(args.eggnog)}): reset to {len(args.eggnog)}.")
            args.cores = len(args.eggnog)
    else:
        logger.error("No valid functional annotations provided: please use '--keggorg' or '--eggnog'.")
        return 1
    
        
    
    # check compatibility of input parameters:
    if args.cnps == '-' and args.biolog != '-':
        logger.error("Missing starting C/N/P/S sources: --biolog must be used in conjunction with --cnps.")
        return 1
    
    
    # get dbexp (dbuni not really used):
    logger.info("Downloading gsrap experimental data...")
    response = get_databases(logger)
    if type(response)==int: return 1
    else: dbuni, dbexp, lastmap = response 
    
    
    # load universe:
    universe = load_input_universe(logger, args.universe)
    if type(universe)==int:
        return 1
    
    

    # disable logging (swith to txt) if strains are more than 1:
    if annotation_source == 'keggorg':
        multistrain = False
    elif annotation_source == 'eggnog':
        multistrain = len(args.eggnog) > 1
        if multistrain:
            logger.info(f"Number of provided strains is >1: logging will be disabled.")
            logger.info(f"Performing {len(args.eggnog)} reconstructions relying on {args.cores} cores... ")
            # actualy this is done inside child processess!
            
            
    # create strain-specific GSMMs using multi-core
    error_raised = False
    sheets_dicts = []
    executor =  confu.ProcessPoolExecutor(max_workers=args.cores)
    
    if annotation_source == 'keggorg':
        futures = [executor.submit(create_model_incore, (annotation_source, universe, None, dbexp, args, multistrain))]
    elif annotation_source == 'eggnog':
        futures = [executor.submit(create_model_incore, (annotation_source, universe, eggpath, dbexp, args, multistrain)) for eggpath in args.eggnog]
        
    for f in confu.as_completed(futures):
        sheets_dict = f.result() 
        
        # error raised in one of the cores:
        if type(sheets_dict) == int: 
            error_raised = True
            break   # a single error from a child process is sufficient to shutdown all the executor

        # this strain reached the end:
        else:
            if multistrain:
                sheets_dicts.append(sheets_dict)
                print(f"{len(sheets_dicts)}/{len(args.eggnog)} ({int(len(sheets_dicts)/len(args.eggnog)*100)}%) completed!", end='\r', file=sys.stderr)
    
    
    # hide last progress trace ('sheets_dicts' unused if not in multi-strain mode):
    if multistrain and sheets_dicts != []:
        last_trace = f"{len(sheets_dicts)}/{len(args.eggnog)} ({int(len(sheets_dicts)/len(args.eggnog)*100)}%) completed!"
        whitewash = ''.join([' ' for i in range(len(last_trace))])
        print(whitewash, end='\r', file=sys.stderr)   
    
    
    # multiproces part terminated: safely shut down the executor
    executor.shutdown(wait=True)
    
    # propagate errors if any:
    if error_raised:
        return 1
    
        
            
    # everything went fine, and there are > 1 strains: time to make the comparison table:
    if multistrain:
        
        logger.info("Collecting comparative tables...")
        comparative_table(logger, args.outdir, sheets_dicts)
            
        
        
    return 0


import argparse
import sys
import traceback
import requests
import importlib.metadata
from datetime import datetime
from packaging import version
import atexit
import os



import cobra


from .commons import get_logger
from .commons import set_usual_formatter
from .commons import set_header_trailer_formatter

from .getmaps import getmaps_command
from .parsedb import parsedb_command
from .mkmodel import mkmodel_command
from .runsims import runsims_command



cobra_config = cobra.Configuration()
solver_name = str(cobra_config.solver.log).split(' ')[1]
solver_name = solver_name.replace("optlang.", '')
solver_name = solver_name.replace("_interface", '')



def main():
    
    
    # define the header of main- and sub-commands.
    current_version = importlib.metadata.metadata("gsrap")["Version"]
    header = f'gsrap v{current_version},\ndeveloped by Gioele Lazzari (gioele.lazzari@univr.it).'
    
    
    # create the command line arguments:
    parser = argparse.ArgumentParser(description=header, add_help=False)
    parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    subparsers = parser.add_subparsers(title='gsrap subcommands', dest='subcommand', help='', required=True)
    
    
    # create the 3 subparsers:
    getmaps_parser = subparsers.add_parser('getmaps', description=header, help="Create 'gsrap.maps' by retrieving metabolic information from specialized resources.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    parsedb_parser = subparsers.add_parser('parsedb', description=header, help="Parse our online database to create the universe GSMM. Requires 'gsrap.maps'.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    mkmodel_parser = subparsers.add_parser('mkmodel', description=header, help='Reconstruct a strain-specific GSMM starting from the universe and a functional annotation.', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    runsims_parser = subparsers.add_parser('runsims', description=header, help='Analyze a GSMM previously built.', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    
    # add arguments for the 'getmaps' command
    getmaps_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    getmaps_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    getmaps_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    getmaps_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    getmaps_parser.add_argument("--keeptmp", action='store_true', help="Do not delete temporary raw files.")
    getmaps_parser.add_argument("--usecache", action='store_true', help="Do not update previously downloaded records (save time).")
    
    
    # add arguments for the 'parsedb' command
    parsedb_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    parsedb_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    parsedb_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    parsedb_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    parsedb_parser.add_argument("-i", "--inmaps", metavar='', type=str, default='./gsrap.maps', help="Input file 'gsrap.maps' previously produced using the 'getmaps' subcommand.")
    parsedb_parser.add_argument("-p", "--progress", action='store_true', help="Show progress for each map.")
    parsedb_parser.add_argument("--module", action='store_true', help="Show progress for each module of each map (use only with --progress).")
    parsedb_parser.add_argument("-f", "--focus", metavar='', type=str, default='-', help="Focus on a particular map/module (use only with --progress).")
    parsedb_parser.add_argument("-m", "--media", metavar='', type=str, default='M9,M9an,M9photo', help="Media to use during growth simulations (comma-separated IDs).")
    parsedb_parser.add_argument("-z", "--initialize", metavar='', type=str, default='-', help="Initialize the universe on the provided medium. By default, the first medium in --media is used. Use 'none' to avoid initialization.")
    parsedb_parser.add_argument("--precursors", action='store_true', help="Verify biosynthesis of biomass precursors and show blocked ones.")
    parsedb_parser.add_argument("--biosynth", action='store_true', help="Check biosynthesis of all metabolites and detect dead-ends.")
    parsedb_parser.add_argument("-t", "--taxon", metavar='', type=str, default='-', help="High-level taxon of interest. If provided, it must follow the syntax '{level}:{name}', where {level} is 'kingdom' or 'phylum'.")
    parsedb_parser.add_argument("-e", "--eggnog", nargs='+', metavar='', type=str, default='-', help="Path to the optional eggnog-mapper annotation table(s).")
    parsedb_parser.add_argument("-k", "--keggorg", metavar='', type=str, default='-', help="A single KEGG Organism code. If provided, it takes precedence over --eggnog.")
    parsedb_parser.add_argument("--goodbefore", metavar='', type=str, default='-', help="Syntax is {pure_mid}-{rid1}-{rid2}. From top to bottom, build the universe until reaction {rid1}, transport {rid2} and metabolite {pure_mid} are reached.")
    parsedb_parser.add_argument("--onlyauthor", metavar='', type=str, default='-', help="Build the universe by parsing contents of the specified author ID only. Contents affected by --goodbefore are parsed anyway.")
    parsedb_parser.add_argument("--nofigs", action='store_true', help="Do not generate figures.")
    parsedb_parser.add_argument("-j", "--justparse", action='store_true', help="Just parse the database without performing extra activities (saves time during universe expansion).")
    
    
    
    # add arguments for the 'mkmodel' command
    mkmodel_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    mkmodel_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    mkmodel_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    mkmodel_parser.add_argument("-c", "--cores", metavar='', type=int, default=0, help="Number of cores to use (if 0, use all available cores).")
    mkmodel_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    mkmodel_parser.add_argument("-e", "--eggnog", nargs='+', metavar='', type=str, default='-', help="Path to the eggnog-mapper annotation table(s).")
    mkmodel_parser.add_argument("-k", "--keggorg", metavar='', type=str, default='-', help="A single KEGG Organism code. If provided, it takes precedence over --eggnog.")
    mkmodel_parser.add_argument("-u", "--universe", metavar='', type=str, default='-', help="Path to the universe model (SBML format).")
    mkmodel_parser.add_argument("-i", "--include", metavar='', type=str, default='-', help="Force the inclusion of the specified reactions (comma-separated IDs).")
    mkmodel_parser.add_argument("-f", "--gapfill", metavar='', type=str, default='-', help="Media to use during gap-filling (comma-separated IDs); if not provided, gap-filling will be skipped.")
    mkmodel_parser.add_argument("-z", "--initialize", metavar='', type=str, default='-', help="Initialize the model on the provided medium. By default, the first medium in --gapfill is used. Use 'none' to avoid initialization.")
    mkmodel_parser.add_argument("-x", "--excludeorp", action='store_true', help="Exclude orphan reactions from the gap-filling repository.")
    #mkmodel_parser.add_argument("-r", "--remove", metavar='', type=str, default='-', help="Force the removal of the specified reactions (comma-separated IDs) (it applies after gap-filling, before Biolog(R)-based curation).")
    mkmodel_parser.add_argument("-l", "--biolog", metavar='', type=str, default='-', help="Strain ID associated to binary Biolog(R) PM1, PM2A, PM3B and PM4A plates; if not provided, Biolog(R)-based model curation will be skipped (use with --cnps and --gap_fill).")
    mkmodel_parser.add_argument("-s", "--cnps", metavar='', type=str, default='glc__D,nh4,pi,so4', help="Starting C, N, P and S source metabolites (comma-separated IDs).")
    mkmodel_parser.add_argument("--conditional", metavar='', type=float, default=0.5, help="Expected minimum fraction of reactions in a biosynthetic pathway for an actually present conditional biomass precursor.")
    mkmodel_parser.add_argument("--biosynth", action='store_true', help="Check biosynthesis of all metabolites and detect dead-ends.")
    mkmodel_parser.add_argument("-b", "--biomass", metavar='', type=str, default='-', help="Strain ID associated to experimental biomass data.")
    mkmodel_parser.add_argument("--nofigs", action='store_true', help="Do not generate figures.")
    
    
    # add arguments for the 'runsims' command
    runsims_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    runsims_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    runsims_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    runsims_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    runsims_parser.add_argument("-i", "--inmodel", metavar='', type=str, default='-', help="Input strain-specific GSMM.")
    runsims_parser.add_argument("-m", "--media", metavar='', type=str, default='-', help="Media to use during growth simulations (comma-separated IDs); if not provided, growth simulations will be skipped.")
    runsims_parser.add_argument("--fva", action='store_true', help="Perform FVA during growth simulations.")
    runsims_parser.add_argument("--biosynth", action='store_true', help="Check biosynthesis of all metabolites and detect dead-ends.")
    runsims_parser.add_argument("--cnps", metavar='', type=str, default='-', help="Starting C, N, P and S source metabolites (comma-separated IDs); if not provided, anternative substrate analysis will be skipped (use with --media).")
    runsims_parser.add_argument("--biolog", metavar='', type=str, default='-', help="Strain ID associated to binary Biolog(R) PM1, PM2A, PM3B and PM4A plates; if not provided, Biolog(R) simulations will be skipped (use with --cnps).")
    runsims_parser.add_argument("--omission", action='store_true', help="Perform single omission experiments to study auxotrophies.")
    runsims_parser.add_argument("--essential", action='store_true', help="Predict essential genes (single-gene knock-out simulations).")
    runsims_parser.add_argument("--factors", action='store_true', help="Predict putative growth factors.")
    runsims_parser.add_argument("--nofigs", action='store_true', help="Do not generate figures.")

    
    # check the inputted subcommand, automatic sys.exit(1) if a bad subprogram was specied. 
    args = parser.parse_args()
    
    
    
    # set up the logger:
    logger = get_logger('gsrap', args.verbose)
    
        
    
    # show a welcome message:
    set_header_trailer_formatter(logger.handlers[0])
    logger.info(header + '\n')
    
    
    
    # check if newer version is available
    try:
        response = requests.get(f"https://pypi.org/pypi/gsrap/json", timeout=3)  # sends an HTTP GET request to the given URL
        response.raise_for_status()  # check the HTTP status code (e.g. 200, 404, 500): if not in the 2xx success range, raise requests.exceptions.HTTPError
        data = response.json()
        newest_version = data["info"]["version"] 
    except Exception as error:  # eg requests.exceptions.Timeout, requests.exceptions.HTTPError
        logger.info(f'Can\'t retrieve the number of the newest version. Please contact the developer reporting the following error: "{error}".')
        logger.info('')  # still no formatting here
        # do not exit, continue with the program
    if version.parse(current_version) < version.parse(newest_version):
        warning_message = f"███ Last version is v{newest_version} and you have v{current_version}: please update gsrap! ███"
        border = ''.join(['█' for i in range(len(warning_message))])
        logger.info(border)
        logger.info(warning_message)
        logger.info(border)
        logger.info('')  # still no formatting here
    
    
    
    # print the full command line:
    command_line = '' 
    for arg, value in vars(args).items():
        if arg == 'subcommand': command_line = command_line + f"gsrap {value} "
        else: command_line = command_line + f"--{arg} {value} "
    logger.info('Inputted command line: "' + command_line.rstrip() + '".\n')
    
    
    
    # The following chunk suppresses the warning 
    # "sys:1: DeprecationWarning: builtin type swigvarlink has no __module__ attribute"
    # raised at Gsrap shutdown by calling memote.suite.api.test_model() in common/memoteutils.py
    def _suppress_swigvarlink_warning():
        sys.stderr = open(os.devnull, 'w')  # tested also with sys.stdout: same effect.
    atexit.register(_suppress_swigvarlink_warning)
    
    
    
    # run the program:
    set_usual_formatter(logger.handlers[0])
    current_date_time = datetime.now()
    formatted_date = current_date_time.strftime("%Y-%m-%d")
    logger.info(f"Welcome to 'gsrap {args.subcommand}', it's {formatted_date}!")
    logger.debug(f"COBRApy started with solver: '{solver_name}'.")
    try: 
        # choose which subcommand to lauch: 
        if args.subcommand == 'getmaps':
            response = getmaps_command(args, logger)
        if args.subcommand == 'parsedb':
            response = parsedb_command(args, logger)
        if args.subcommand == 'mkmodel':
            response = mkmodel_command(args, logger)
        if args.subcommand == 'runsims':
            response = runsims_command(args, logger)
            
        if response == 0:
            logger.info(f"'gsrap {args.subcommand}' terminated without errors!")
    except: 
        # show the error stack trace for this un-handled error: 
        response = 1
        logger.error('Traceback is reported below.\n\n' + traceback.format_exc())


    
    # terminate the program:
    set_header_trailer_formatter(logger.handlers[0])
    if response == 1: 
        print(file=sys.stderr)  # separate last error from fresh prompt
        sys.exit(1)
    else: 
        # show a bye message
        logger.info('\n' + header)
        sys.exit(0) # exit without errors
        
        
        
if __name__ == "__main__":
    main()


import warnings

import pandas as pnd
import gempipe


from ..commons import get_biomass_dict
from ..commons import apply_medium_given_column
from ..commons import verify_growth
from ..commons import get_biomass_df_structure
from ..commons import fba_no_warnings

from .gapfillutils import get_repository_nogenes
from .gapfillutils import import_from_universe



def gapfill_on_media(logger, model, universe, dbexp, media, cond_col_dict, exclude_orphans):
    
    
    # check if requested (always)
    df_B = pnd.DataFrame()
    df_B = get_biomass_df_structure(logger, df_B, model, universe, dbexp, cond_col_dict)
         
    
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: gap-filling will be skipped.")
        return df_B
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in dbexp['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(dbexp['media'].columns)}.")
        return 1
        
        
    # get the repository of reactions:
    repository_nogenes = get_repository_nogenes(logger, universe, exclude_orphans)
    
    
    logger.info(f"Gap-filling for biomass on {len(media)} media...")
    for medium in media: 
        if medium not in dbexp['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_B[f'{medium}'] = None
        logger.debug(f"Gap-filling on medium '{medium}'...")
                    
            
        # apply medium both on universe and model:
        response = apply_medium_given_column(logger, repository_nogenes, medium, dbexp['media'][medium])
        if response == 1: return 1
        if not verify_growth(repository_nogenes):
            
            logger.error(f"Medium '{medium}' does not support growth of universe.")
            return 1
        
        response = apply_medium_given_column(logger, model, medium, dbexp['media'][medium])
        if response == 1: return 1
        if verify_growth(model):
            
            logger.debug(f"No need to gapfill model on medium '{medium}'.")
            # show result of vanilla FBA if --verbose:
            _, plain_objval, plain_status = fba_no_warnings(model)
            logger.debug(f"Plain FBA on medium '{medium}': {plain_objval} ({plain_status}).")
            continue


        # launch gap-filling separately for each biomass precursor:
        # force 'atp_c' as first:
        universal_precs = ['atp_c'] + [m.id for m in universe.reactions.Biomass.reactants if m.id not in ['atp_c', 'h2o_c']]
        for mid in universal_precs:
            # skip conditional precursors: 
            if mid in [m.id for m in model.reactions.Biomass.reactants]:
            
                # save time if it can already be synthesized
                if gempipe.can_synth(model, mid)[0]:
                    df_B.loc[mid, f'{medium}'] = '/'
                    logger.debug(f"Gap-filled 0 reactions on medium '{medium}' for '{mid}': [].")
                    continue   # save time!
                    
                    
                # remove 'meanfa_c' sink (needed to model fatty acids / phospholips):
                with repository_nogenes, model:
                    
                    # needed to avoid some infeasible solutions (its' temporary thanks to the 'with' statement)
                    repository_nogenes.reactions.Biomass.bounds = (0,0)
                    model.reactions.Biomass.bounds = (0,0)
                    
                    with warnings.catch_warnings():
                        # avoid warnings like "python3.9/site-packages/cobra/core/group.py:147: UserWarning: need to pass in a list"
                        warnings.simplefilter("ignore")
                        repository_nogenes.remove_reactions(["sn_meanfa_c"]) # this works also with rids instaead of Rs

                    #minflux = get_optthr()
                    minflux = 0.1
                    suggested_rids = gempipe.perform_gapfilling(model, repository_nogenes, mid, nsol=1, minflux=minflux, boost=True, verbose=False)
                    if suggested_rids == None:
                        logger.error(f"The gap-filling problem seems too hard for '{mid}' on medium '{medium}': should '{mid}' be included in '{medium}'?")
                        return 1
                
                
                df_B.loc[mid, f'{medium}'] = '; '.join([f'+{i}' for i in suggested_rids]) 
                logger.debug(f"Gap-filled {len(suggested_rids)} reactions on medium '{medium}' for '{mid}': {suggested_rids}.")
                for rid in suggested_rids:
                    import_from_universe(model, repository_nogenes, rid, gpr='')
                    
                    
        # if reached this point , the model should be ble to grow on this medium
        _, plain_objval, plain_status = fba_no_warnings(model)
        logger.debug(f"Plain FBA on medium '{medium}': {plain_objval} ({plain_status}).")
                
                
    return df_B

